#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import dataclasses
import warnings
from contextlib import ExitStack
from copy import deepcopy
from itertools import product
from unittest import mock
from unittest.mock import Mock, patch

import numpy as np
import torch
from ax.core.search_space import SearchSpaceDigest
from ax.exceptions.model import ModelError
from ax.generators.torch.botorch_modular.acquisition import Acquisition
from ax.generators.torch.botorch_modular.generator import BoTorchGenerator
from ax.generators.torch.botorch_modular.kernels import ScaleMaternKernel
from ax.generators.torch.botorch_modular.multi_acquisition import MultiAcquisition
from ax.generators.torch.botorch_modular.surrogate import Surrogate, SurrogateSpec
from ax.generators.torch.botorch_modular.utils import (
    choose_botorch_acqf_class,
    choose_model_class,
    construct_acquisition_and_optimizer_options,
    fit_botorch_model,
    ModelConfig,
)
from ax.generators.torch.utils import _filter_X_observed, predict_from_model
from ax.generators.torch_base import TorchOptConfig
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.testing.mock import mock_botorch_optimize
from ax.utils.testing.torch_stubs import get_torch_test_data
from botorch.acquisition import qLogNoisyExpectedImprovement
from botorch.acquisition.analytic import PosteriorMean
from botorch.acquisition.input_constructors import (
    _register_acqf_input_constructor,
    get_acqf_input_constructor,
)
from botorch.acquisition.logei import qLogProbabilityOfFeasibility
from botorch.acquisition.monte_carlo import qExpectedImprovement
from botorch.acquisition.multi_objective.logei import (
    qLogNoisyExpectedHypervolumeImprovement,
)
from botorch.acquisition.multi_objective.objective import WeightedMCMultiOutputObjective
from botorch.acquisition.objective import (
    GenericMCObjective,
    ScalarizedPosteriorTransform,
)
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.models.gp_regression import SingleTaskGP
from botorch.models.gp_regression_fidelity import SingleTaskMultiFidelityGP
from botorch.models.model import Model, ModelList
from botorch.sampling.normal import SobolQMCNormalSampler
from botorch.utils.constraints import get_outcome_constraint_transforms
from botorch.utils.datasets import SupervisedDataset
from botorch.utils.testing import skip_if_import_error
from botorch.utils.types import DEFAULT
from gpytorch.likelihoods.gaussian_likelihood import FixedNoiseGaussianLikelihood
from pyre_extensions import assert_is_instance, none_throws


CURRENT_PATH: str = __name__
MODEL_PATH: str = BoTorchGenerator.__module__
SURROGATE_PATH: str = Surrogate.__module__
ACQUISITION_PATH: str = Acquisition.__module__
UTILS_PATH = f"{ModelConfig.__module__}"

ACQ_OPTIONS: dict[str, SobolQMCNormalSampler] = {
    Keys.SAMPLER: SobolQMCNormalSampler(sample_shape=torch.Size([1024]))
}


class BoTorchGeneratorTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.surrogate = Surrogate()
        self.acquisition_class = Acquisition
        self.botorch_acqf_class = qExpectedImprovement
        self.botorch_acqf_options = ACQ_OPTIONS
        self.model = BoTorchGenerator(
            surrogate=self.surrogate,
            acquisition_class=self.acquisition_class,
            botorch_acqf_class=self.botorch_acqf_class,
            botorch_acqf_options=self.botorch_acqf_options,
        )

        self.dtype = torch.float
        self.device = torch.device("cpu")
        tkwargs = {"dtype": self.dtype, "device": self.device}
        self.tkwargs = tkwargs
        (
            self.Xs,
            self.Ys,
            self.Yvars,
            self.bounds,
            _,
            self.feature_names,  # This is ["x1", "x2", "x3"].
            self.metric_names,  # This is just ["y"].
        ) = get_torch_test_data(dtype=self.dtype)
        Xs2, Ys2, Yvars2, _, _, _, _ = get_torch_test_data(dtype=self.dtype, offset=1.0)
        self.X_test = Xs2
        self.block_design_training_data = [
            SupervisedDataset(
                X=self.Xs,
                Y=self.Ys,
                Yvar=self.Yvars,
                feature_names=self.feature_names,
                outcome_names=self.metric_names,
            )
        ]
        self.non_block_design_training_data = self.block_design_training_data + [
            SupervisedDataset(
                X=Xs2,
                Y=Ys2,
                Yvar=Yvars2,
                feature_names=self.feature_names,
                outcome_names=["y2"],
            )
        ]
        self.search_space_digest = SearchSpaceDigest(
            feature_names=self.feature_names,
            bounds=[(0.0, 10.0), (0.0, 10.0), (0.0, 10.0)],
        )
        self.mf_search_space_digest = SearchSpaceDigest(
            feature_names=self.feature_names,
            bounds=[(0.0, 10.0), (0.0, 10.0), (0.0, 10.0)],
            fidelity_features=[2],
            target_values={1: 1.0},
        )
        self.candidate_metadata = []
        self.optimizer_options = {Keys.NUM_RESTARTS: 40, Keys.RAW_SAMPLES: 1024}
        self.model_gen_options = {
            Keys.OPTIMIZER_KWARGS: self.optimizer_options,
            Keys.ACQF_KWARGS: {"eta": 3.0},
            Keys.AX_ACQUISITION_KWARGS: {Keys.SUBSET_MODEL: False},
        }
        self.objective_weights = torch.tensor([1.0], **tkwargs)
        self.outcome_constraints = (
            torch.tensor([[1.0]], **tkwargs),
            torch.tensor([[3.5]], **tkwargs),
        )
        self.moo_objective_weights = torch.tensor([1.0, 1.5, 0.0], **tkwargs)
        self.moo_objective_thresholds = torch.tensor(
            [0.5, 1.5, float("nan")], **tkwargs
        )
        self.moo_outcome_constraints = (
            torch.tensor([[1.0, 0.0, 0.0]], **tkwargs),
            torch.tensor([[3.5]], **tkwargs),
        )
        self.linear_constraints = None
        self.fixed_features = None
        self.pending_observations = None
        self.moo_metric_names = ["y1", "y2", "y3"]
        self.moo_training_data = [  # block design
            SupervisedDataset(
                X=X,
                Y=Y,
                Yvar=Yvar,
                feature_names=self.feature_names,
                outcome_names=[mn],
            )
            for X, Y, Yvar, mn in zip(
                [self.Xs for _ in range(3)],
                [self.Ys, Ys2, self.Ys],
                [self.Yvars for _ in range(3)],
                self.moo_metric_names,
            )
        ]

        self.torch_opt_config = TorchOptConfig(
            objective_weights=self.objective_weights,
            outcome_constraints=self.outcome_constraints,
            linear_constraints=self.linear_constraints,
            fixed_features=self.fixed_features,
            pending_observations=self.pending_observations,
            model_gen_options=self.model_gen_options,
        )
        self.moo_torch_opt_config = dataclasses.replace(
            self.torch_opt_config,
            objective_weights=self.moo_objective_weights,
            objective_thresholds=self.moo_objective_thresholds,
            outcome_constraints=self.moo_outcome_constraints,
            is_moo=True,
            model_gen_options={
                Keys.OPTIMIZER_KWARGS: self.optimizer_options,
                Keys.ACQF_KWARGS: {"eta": 3.0},
                Keys.AX_ACQUISITION_KWARGS: {Keys.SUBSET_MODEL: True},
            },
        )
        self.torch_opt_config_infeas = dataclasses.replace(
            self.torch_opt_config,
            outcome_constraints=(
                torch.tensor([[1.0]], **tkwargs),
                torch.tensor([[-5.0]], **tkwargs),
            ),
        )
        self.moo_torch_opt_config_infeas = dataclasses.replace(
            self.moo_torch_opt_config,
            outcome_constraints=(
                torch.tensor([[1.0]], **tkwargs),
                torch.tensor([[-5.0]], **tkwargs),
            ),
        )

    def test_init(self) -> None:
        # Default model with no specifications.
        model = BoTorchGenerator()
        self.assertEqual(model.acquisition_class, Acquisition)
        # Model that specifies `botorch_acqf_class`.
        model = BoTorchGenerator(botorch_acqf_class=qExpectedImprovement)
        self.assertEqual(model.acquisition_class, Acquisition)
        self.assertEqual(model.botorch_acqf_class, qExpectedImprovement)

        # Check defaults for refitting settings.
        self.assertFalse(model.refit_on_cv)
        self.assertTrue(model.warm_start_refit)

        # Check setting non-default refitting settings
        mdl2 = BoTorchGenerator(
            surrogate=self.surrogate,
            acquisition_class=self.acquisition_class,
            botorch_acqf_options=self.botorch_acqf_options,
            refit_on_cv=True,
            warm_start_refit=False,
        )
        self.assertTrue(mdl2.refit_on_cv)
        self.assertFalse(mdl2.warm_start_refit)

    def test_surrogate_property(self) -> None:
        self.assertIs(self.surrogate, self.model.surrogate)

    def test_Xs_property(self) -> None:
        self.model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )

        self.assertEqual(len(self.model.Xs), 1)
        self.assertTrue(
            self.model.Xs[0].equal(torch.tensor([[1.0, 2.0, 3.0], [2.0, 3.0, 4.0]]))
        )

    def test_dtype(self) -> None:
        self.model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        self.assertEqual(self.model.dtype, torch.float32)

    def test_device(self) -> None:
        self.model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        self.assertEqual(self.model.device, torch.device("cpu"))

    def test_botorch_acqf_class_property(self) -> None:
        self.assertEqual(self.botorch_acqf_class, self.model.botorch_acqf_class)
        self.model._botorch_acqf_class = None
        with self.assertRaisesRegex(
            ValueError, "BoTorch `AcquisitionFunction` has not yet been set."
        ):
            self.model.botorch_acqf_class

    def test__construct(self) -> None:
        """Test autoset."""
        self.model._surrogate = None
        with mock.patch(
            f"{UTILS_PATH}.choose_model_class", wraps=choose_model_class
        ) as mock_choose_model_class:
            self.model.fit(
                datasets=self.block_design_training_data,
                search_space_digest=self.mf_search_space_digest,
                candidate_metadata=self.candidate_metadata,
            )
        # `choose_model_class` is called.
        mock_choose_model_class.assert_called_with(
            dataset=self.block_design_training_data[0],
            search_space_digest=self.mf_search_space_digest,
        )

    # This mock is hard to remove since it is mocks a method on a surrogate that
    # is only constructed when `model.fit` is called
    @mock.patch(f"{SURROGATE_PATH}.Surrogate._construct_model")
    def test_fit(self, mock_fit: Mock) -> None:
        # If surrogate is not yet set, initialize it internally.
        self.model._surrogate = None
        with self.assertRaisesRegex(
            ModelError, "Surrogate has not yet been constructed."
        ):
            self.model.search_space_digest  # can't access before fit

        self.model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.mf_search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )

        self.assertIsInstance(self.model.search_space_digest, SearchSpaceDigest)
        self.assertEqual(self.model.search_space_digest, self.mf_search_space_digest)

        # Since we want to refit on updates but not warm start refit, we clear the
        # state dict.
        mock_fit.assert_called_with(
            dataset=self.block_design_training_data[0],
            search_space_digest=self.mf_search_space_digest,
            model_config=ModelConfig(
                botorch_model_class=None,
                model_options={},
                mll_class=None,
                mll_options={},
                input_transform_classes=DEFAULT,
                input_transform_options={},
                outcome_transform_classes=None,
                outcome_transform_options={},
                covar_module_class=None,
                covar_module_options={},
                likelihood_class=None,
                likelihood_options={},
                name="default",
            ),
            state_dict=None,
            refit=True,
        )

    def test_predict(self) -> None:
        with mock.patch.object(self.model._surrogate, "predict") as mock_predict:
            self.model.predict(X=self.X_test)
        mock_predict.assert_called_with(X=self.X_test, use_posterior_predictive=False)
        with mock.patch.object(self.model._surrogate, "predict") as mock_predict:
            self.model.predict(X=self.X_test, use_posterior_predictive=True)
        mock_predict.assert_called_with(X=self.X_test, use_posterior_predictive=True)

    @mock_botorch_optimize
    def test_cross_validate(self) -> None:
        self.model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )

        old_surrogate = self.model.surrogate
        clone_surrogate = old_surrogate.clone_reset()

        for refit_on_cv, warm_start_refit, use_posterior_predictive in product(
            (True, False), (True, False), (True, False)
        ):
            self.model.refit_on_cv = refit_on_cv
            self.model.warm_start_refit = warm_start_refit
            with ExitStack() as es:
                mock_clone = es.enter_context(
                    mock.patch.object(
                        old_surrogate, "clone_reset", return_value=clone_surrogate
                    )
                )
                mock_fit = es.enter_context(
                    mock.patch.object(clone_surrogate, "fit", wraps=clone_surrogate.fit)
                )
                mock_predict_orig_surrogate = es.enter_context(
                    mock.patch.object(
                        self.model.surrogate,
                        "predict",
                        wraps=self.model.surrogate.predict,
                    )
                )
                mock_predict_any_surrogate = es.enter_context(
                    mock.patch(
                        f"{SURROGATE_PATH}.predict_from_model", wraps=predict_from_model
                    )
                )
                self.model.cross_validate(
                    datasets=self.block_design_training_data,
                    X_test=self.X_test,
                    search_space_digest=self.search_space_digest,
                    use_posterior_predictive=use_posterior_predictive,
                )
            mock_clone.assert_called_once()
            # Check that `predict` is called on the cloned surrogate, not
            # on the original one.
            mock_predict_orig_surrogate.assert_not_called()
            mock_predict_any_surrogate.assert_called_once()

            # Check correct X_test.
            kwargs = mock_predict_any_surrogate.call_args.kwargs
            self.assertTrue(torch.equal(kwargs["X"], self.X_test))
            self.assertIs(kwargs["use_posterior_predictive"], use_posterior_predictive)

            # Check that surrogate is reset back to `old_surrogate` at the
            # end of cross-validation.
            self.assertIs(self.model.surrogate, old_surrogate)

            expected_state_dict = (
                None
                if refit_on_cv and not warm_start_refit
                else self.model.surrogate.model.state_dict()
            )

            # Check correct `refit` and `state_dict` values.
            kwargs = mock_fit.call_args.kwargs
            self.assertEqual(kwargs["refit"], refit_on_cv)
            if expected_state_dict is None:
                self.assertIsNone(kwargs["state_dict"], expected_state_dict)
            else:
                self.assertEqual(
                    kwargs["state_dict"].keys(), expected_state_dict.keys()
                )

    @mock_botorch_optimize
    def test_cross_validate_multiple_configs(self) -> None:
        """Test cross-validation with multiple configs."""
        for refit_on_cv in (True, False):
            with self.subTest(refit_on_cv=refit_on_cv):
                self.model = BoTorchGenerator(
                    surrogate_spec=SurrogateSpec(
                        model_configs=[
                            ModelConfig(),
                            ModelConfig(
                                botorch_model_class=SingleTaskGP,
                                covar_module_class=ScaleMaternKernel,
                            ),
                        ]
                    ),
                    acquisition_class=self.acquisition_class,
                    botorch_acqf_class=self.botorch_acqf_class,
                    botorch_acqf_options=self.botorch_acqf_options,
                    refit_on_cv=refit_on_cv,
                )
                self.model.fit(
                    datasets=self.block_design_training_data,
                    search_space_digest=self.search_space_digest,
                    candidate_metadata=self.candidate_metadata,
                )
                with patch(
                    f"{Surrogate.__module__}.fit_botorch_model", wraps=fit_botorch_model
                ) as mock_fit:
                    self.model.cross_validate(
                        datasets=self.block_design_training_data,
                        X_test=self.X_test,
                        search_space_digest=self.search_space_digest,
                    )
                # check that we don't fit the model during cross_validation
                if refit_on_cv:
                    mock_fit.assert_called()
                else:
                    mock_fit.assert_not_called()

    @mock_botorch_optimize
    @mock.patch(
        f"{MODEL_PATH}.construct_acquisition_and_optimizer_options",
        wraps=construct_acquisition_and_optimizer_options,
    )
    @mock.patch(
        f"{MODEL_PATH}.choose_botorch_acqf_class", wraps=choose_botorch_acqf_class
    )
    def _test_gen(
        self,
        mock_choose_botorch_acqf_class: Mock,
        mock_construct_options: Mock,
        botorch_model_class: type[Model],
        search_space_digest: SearchSpaceDigest,
    ) -> None:
        qLogNEI_input_constructor = get_acqf_input_constructor(
            qLogNoisyExpectedImprovement
        )
        mock_input_constructor = mock.MagicMock(
            qLogNEI_input_constructor, side_effect=qLogNEI_input_constructor
        )
        _register_acqf_input_constructor(
            acqf_cls=qLogNoisyExpectedImprovement,
            input_constructor=mock_input_constructor,
        )
        mock_optimize_return_value = (
            torch.tensor([[1.0]]),
            torch.tensor([2.0]),
            torch.tensor([1.0]),
        )
        surrogate = Surrogate(
            surrogate_spec=SurrogateSpec(
                model_configs=[ModelConfig(botorch_model_class=botorch_model_class)]
            )
        )
        model = BoTorchGenerator(
            surrogate=surrogate,
            acquisition_class=Acquisition,
            botorch_acqf_options=self.botorch_acqf_options,
        )
        # Assert that error is raised if we haven't fit the model
        with self.assertRaisesRegex(ModelError, "fit the model first"):
            model.gen(
                n=1,
                search_space_digest=search_space_digest,
                torch_opt_config=self.torch_opt_config,
            )
        model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=search_space_digest,
        )
        with ExitStack() as es:
            mock_init_acqf = es.enter_context(
                mock.patch.object(
                    BoTorchGenerator,
                    "_instantiate_acquisition",
                    wraps=model._instantiate_acquisition,
                )
            )
            mock_optimize = es.enter_context(
                mock.patch(
                    f"{CURRENT_PATH}.Acquisition.optimize",
                    return_value=mock_optimize_return_value,
                )
            )

            gen_results = model.gen(
                n=1,
                search_space_digest=search_space_digest,
                torch_opt_config=self.torch_opt_config,
            )
        # Assert acquisition initialized with expected arguments
        mock_init_acqf.assert_called_once_with(
            search_space_digest=search_space_digest,
            torch_opt_config=self.torch_opt_config,
            botorch_acqf_options={**ACQ_OPTIONS, "eta": 3.0},
            acq_options={Keys.SUBSET_MODEL: False},
            botorch_acqf_classes_with_options=None,
            n=1,
        )

        mock_input_constructor.assert_called_once()
        ckwargs = mock_input_constructor.call_args[1]

        # We particularly want to make sure that args that will not be used are
        # not passed
        expected_kwargs = {
            "bounds",
            "constraints",
            "X_baseline",
            "sampler",
            "objective",
            "training_data",
            "model",
            "eta",
            "constraints_tuple",
        }
        self.assertSetEqual(set(ckwargs.keys()), expected_kwargs)
        for k in expected_kwargs:
            self.assertIsNotNone(ckwargs[k], f"{k} is None")

        m = ckwargs["model"]
        self.assertIsInstance(m, botorch_model_class)
        self.assertEqual(m.num_outputs, 1)
        training_data = ckwargs["training_data"]
        self.assertIsInstance(training_data, SupervisedDataset)
        self.assertTrue(torch.equal(training_data.X, self.Xs))
        self.assertTrue(
            torch.equal(
                training_data.Y,
                torch.cat([ds.Y for ds in self.block_design_training_data], dim=-1),
            )
        )

        self.assertIsInstance(ckwargs["objective"], GenericMCObjective)
        expected_X_baseline = _filter_X_observed(
            Xs=[dataset.X for dataset in self.block_design_training_data],
            objective_weights=self.objective_weights,
            outcome_constraints=self.outcome_constraints,
            bounds=search_space_digest.bounds,
            linear_constraints=self.linear_constraints,
            fixed_features=self.fixed_features,
        )
        self.assertTrue(
            torch.equal(ckwargs["X_baseline"], none_throws(expected_X_baseline))
        )

        mock_construct_options.assert_called_with(
            acqf_options={},
            botorch_acqf_options=self.botorch_acqf_options,
            model_gen_options={
                **self.model_gen_options,
                Keys.AX_ACQUISITION_KWARGS: {Keys.SUBSET_MODEL: False},
            },
            botorch_acqf_classes_with_options=None,
        )
        # Assert `choose_botorch_acqf_class` is called
        mock_choose_botorch_acqf_class.assert_called_once()
        self.assertEqual(model._botorch_acqf_class, qLogNoisyExpectedImprovement)

        # Assert `optimize` called with kwargs
        mock_optimize.assert_called_with(
            n=1,
            search_space_digest=search_space_digest,
            inequality_constraints=None,
            fixed_features=self.fixed_features,
            rounding_func=None,
            optimizer_options=self.optimizer_options,
        )
        # make sure ACQF_KWARGS are passed properly
        self.assertEqual(none_throws(model._acquisition).acqf._eta, 3.0)

        _register_acqf_input_constructor(
            acqf_cls=qLogNoisyExpectedImprovement,
            input_constructor=qLogNEI_input_constructor,
        )

        # Make sure `gen` runs without mocking out Acquisition.optimize
        with self.subTest("No mocks"):
            gen_results = model.gen(
                n=1,
                search_space_digest=search_space_digest,
                torch_opt_config=self.torch_opt_config,
            )
            self.assertTrue(torch.isfinite(gen_results.points).all())

    def test_gen_SingleTaskGP(self) -> None:
        self._test_gen(
            botorch_model_class=SingleTaskGP,
            search_space_digest=self.search_space_digest,
        )

    def test_gen_SingleTaskMultiFidelityGP(self) -> None:
        self._test_gen(
            botorch_model_class=SingleTaskMultiFidelityGP,
            search_space_digest=self.mf_search_space_digest,
        )

    @mock_botorch_optimize
    def test_feature_importances(self) -> None:
        for botorch_model_class in [SingleTaskGP, SaasFullyBayesianSingleTaskGP]:
            surrogate = Surrogate(
                surrogate_spec=SurrogateSpec(
                    model_configs=[ModelConfig(botorch_model_class=botorch_model_class)]
                )
            )
            model = BoTorchGenerator(
                surrogate=surrogate,
                acquisition_class=Acquisition,
                botorch_acqf_options=self.botorch_acqf_options,
            )
            model.surrogate.fit(
                datasets=self.block_design_training_data,
                search_space_digest=self.search_space_digest,
            )
            if botorch_model_class == SaasFullyBayesianSingleTaskGP:
                mcmc_samples = {
                    "lengthscale": torch.tensor(
                        [[1, 2, 3], [2, 3, 4], [3, 4, 5]], **self.tkwargs
                    ),
                    "outputscale": torch.rand(3, **self.tkwargs),
                    "mean": torch.randn(3, **self.tkwargs),
                    "noise": torch.rand(3, **self.tkwargs),
                }
                # pyre-fixme[29]: `Union[Tensor, Module]` is not a function.
                model.surrogate.model.load_mcmc_samples(mcmc_samples)
                importances = model.feature_importances()
                self.assertTrue(
                    np.allclose(importances, np.array([6 / 13, 4 / 13, 3 / 13]))
                )
                self.assertEqual(importances.shape, (1, 1, 3))
                saas_model = deepcopy(model.surrogate.model)
            else:
                # pyre-fixme[16]: `Tensor` has no attribute `lengthscale`.
                # pyre-fixme[16]: `Module` has no attribute `lengthscale`.
                model.surrogate.model.covar_module.lengthscale = torch.tensor(
                    [1, 2, 3], **self.tkwargs
                )
                importances = model.feature_importances()
                self.assertTrue(
                    np.allclose(importances, np.array([6 / 11, 3 / 11, 2 / 11]))
                )
                self.assertEqual(importances.shape, (1, 1, 3))
                vanilla_model = deepcopy(model.surrogate.model)

        # Mixed model
        model.surrogate._model = ModelList(saas_model, vanilla_model)  # pyre-ignore
        importances = model.feature_importances()
        self.assertTrue(
            np.allclose(
                importances,
                np.expand_dims(
                    np.array([[6 / 13, 4 / 13, 3 / 13], [6 / 11, 3 / 11, 2 / 11]]),
                    axis=1,
                ),
            )
        )
        self.assertEqual(importances.shape, (2, 1, 3))
        # Add model we don't support
        vanilla_model.covar_module = None
        model.surrogate._model = vanilla_model  # pyre-ignore
        with self.assertRaisesRegex(
            NotImplementedError,
            "Failed to extract lengthscales from `m.covar_module` "
            "and `m.covar_module.base_kernel`",
        ):
            model.feature_importances()
        # Test model is None
        model.surrogate._model = None
        with self.assertRaisesRegex(
            ModelError, "BoTorch `Model` has not yet been constructed"
        ):
            model.feature_importances()

    @mock_botorch_optimize
    def test_best_point(self) -> None:
        self.model._surrogate = None
        self.model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.mf_search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        self.assertIsNotNone(
            self.model.best_point(
                search_space_digest=self.mf_search_space_digest,
                torch_opt_config=self.torch_opt_config,
            )
        )
        with mock.patch(f"{SURROGATE_PATH}.best_in_sample_point", return_value=None):
            self.assertIsNone(
                self.model.best_point(
                    search_space_digest=self.mf_search_space_digest,
                    torch_opt_config=self.torch_opt_config,
                )
            )
        with self.assertRaisesRegex(NotImplementedError, "Best observed"):
            self.model.best_point(
                search_space_digest=self.mf_search_space_digest,
                torch_opt_config=dataclasses.replace(
                    self.torch_opt_config, is_moo=True
                ),
            )

    @mock_botorch_optimize
    def test_evaluate_acquisition_function(self) -> None:
        model = BoTorchGenerator(
            surrogate=self.surrogate,
            acquisition_class=Acquisition,
            botorch_acqf_options=self.botorch_acqf_options,
        )
        model.surrogate.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.search_space_digest,
        )
        points = model.evaluate_acquisition_function(
            X=self.X_test,
            search_space_digest=self.search_space_digest,
            torch_opt_config=self.torch_opt_config,
            acq_options={},
        )
        self.assertEqual(points.shape, torch.Size([1]))
        # testing that the new setup chooses qLogNEI by default
        self.assertEqual(model._botorch_acqf_class, qLogNoisyExpectedImprovement)

    @mock_botorch_optimize
    @mock.patch(
        f"{MODEL_PATH}.choose_botorch_acqf_class", wraps=choose_botorch_acqf_class
    )
    def test_p_feas(
        self,
        mock_choose_botorch_acqf_class: Mock,
    ) -> None:
        """Test that we dispatch to `qLogProbabilityOfFeasibility`
        when no feasible points have been found"""
        moo_training_data_non_block = [
            SupervisedDataset(
                X=X,
                Y=Y,
                Yvar=Yvar,
                feature_names=self.feature_names,
                outcome_names=[mn],
            )
            for X, Y, Yvar, mn in zip(
                [self.Xs, self.Xs[:1], self.Xs[1:]],
                [self.Ys, self.Ys[:1] + 1, self.Ys[1:]],
                [self.Yvars, self.Yvars[:1], self.Yvars[1:]],
                self.moo_metric_names,
            )
        ]

        for datasets, torch_opt_config in zip(
            (
                self.block_design_training_data,
                self.moo_training_data,
                moo_training_data_non_block,
            ),
            (
                self.torch_opt_config_infeas,
                self.moo_torch_opt_config_infeas,
                self.moo_torch_opt_config_infeas,
            ),
        ):
            model = BoTorchGenerator(
                surrogate=self.surrogate,
                acquisition_class=Acquisition,
                botorch_acqf_options=self.botorch_acqf_options,
            )
            model.surrogate.fit(
                datasets=datasets,
                search_space_digest=self.search_space_digest,
            )
            self.assertIsNone(model._botorch_acqf_class)  # Should not have been set
            # Gen
            with self.subTest("No mocks"):
                with warnings.catch_warnings(record=True) as ws:
                    gen_results = model.gen(
                        n=1,
                        search_space_digest=self.search_space_digest,
                        torch_opt_config=torch_opt_config,
                    )
                self.assertEqual(len(ws), 0)
                mock_choose_botorch_acqf_class.assert_called()
                mock_choose_botorch_acqf_class.reset_mock()
                self.assertEqual(
                    model._botorch_acqf_class, qLogProbabilityOfFeasibility
                )
                self.assertTrue(torch.isfinite(gen_results.points).all())
            # Evaluate acqf
            points = model.evaluate_acquisition_function(
                X=self.X_test,
                search_space_digest=self.search_space_digest,
                torch_opt_config=torch_opt_config,
                acq_options={},
            )
            mock_choose_botorch_acqf_class.assert_called()
            mock_choose_botorch_acqf_class.reset_mock()
            self.assertEqual(points.shape, torch.Size([1]))
            # Setting use_p_feasible to False should turn off pFeas
            model = BoTorchGenerator(
                surrogate=self.surrogate,
                acquisition_class=Acquisition,
                botorch_acqf_options=self.botorch_acqf_options,
                use_p_feasible=False,
            )
            model.surrogate.fit(
                datasets=datasets,
                search_space_digest=self.search_space_digest,
            )
            gen_results = model.gen(
                n=1,
                search_space_digest=self.search_space_digest,
                torch_opt_config=torch_opt_config,
            )
            self.assertEqual(
                model._botorch_acqf_class,
                qLogNoisyExpectedHypervolumeImprovement
                if torch_opt_config.is_moo
                else qLogNoisyExpectedImprovement,
            )

    @mock_botorch_optimize
    def test_surrogate_model_options_propagation(self) -> None:
        surrogate_spec = SurrogateSpec()
        model = BoTorchGenerator(surrogate_spec=surrogate_spec)
        with mock.patch(f"{MODEL_PATH}.Surrogate", wraps=Surrogate) as mock_init:
            model.fit(
                datasets=self.non_block_design_training_data,
                search_space_digest=self.mf_search_space_digest,
                candidate_metadata=self.candidate_metadata,
            )
        mock_init.assert_called_with(
            surrogate_spec=surrogate_spec, refit_on_cv=False, warm_start_refit=True
        )

    @mock_botorch_optimize
    def test_surrogate_options_propagation(self) -> None:
        surrogate_spec = SurrogateSpec(allow_batched_models=False)
        model = BoTorchGenerator(surrogate_spec=surrogate_spec)
        with mock.patch(f"{MODEL_PATH}.Surrogate", wraps=Surrogate) as mock_init:
            model.fit(
                datasets=self.non_block_design_training_data,
                search_space_digest=self.mf_search_space_digest,
                candidate_metadata=self.candidate_metadata,
            )
        mock_init.assert_called_with(
            surrogate_spec=surrogate_spec, refit_on_cv=False, warm_start_refit=True
        )

    @mock_botorch_optimize
    def test_model_list_choice(self) -> None:
        model = BoTorchGenerator()
        model.fit(
            datasets=self.non_block_design_training_data,
            search_space_digest=self.mf_search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        # A model list should be chosen, since Xs are not all the same.
        model_list = assert_is_instance(model.surrogate.model, ModelList)
        for submodel in model_list.models:
            # There are fidelity features and nonempty Yvars, so
            # MFGP should be chosen.
            self.assertIsInstance(submodel, SingleTaskMultiFidelityGP)

    @mock_botorch_optimize
    def test_MOO(self) -> None:
        # Add mock for qLogNEHVI input constructor to catch arguments passed to it.
        qLogNEHVI_input_constructor = get_acqf_input_constructor(
            qLogNoisyExpectedHypervolumeImprovement
        )
        mock_input_constructor = mock.MagicMock(
            qLogNEHVI_input_constructor, side_effect=qLogNEHVI_input_constructor
        )
        _register_acqf_input_constructor(
            acqf_cls=qLogNoisyExpectedHypervolumeImprovement,
            input_constructor=mock_input_constructor,
        )

        model = BoTorchGenerator()
        model.fit(
            datasets=self.moo_training_data,
            search_space_digest=self.search_space_digest,
            candidate_metadata=self.candidate_metadata,
        )
        self.assertIsInstance(model.surrogate.model, SingleTaskGP)
        subset_outcome_constraints = (
            # model is subset since last output is not used
            self.moo_outcome_constraints[0][:, :2],
            self.moo_outcome_constraints[1],
        )
        constraints = get_outcome_constraint_transforms(
            outcome_constraints=subset_outcome_constraints,
        )
        with mock.patch(
            f"{ACQUISITION_PATH}.get_outcome_constraint_transforms",
            # Dummy candidates and acquisition function value.
            # This will return the same value as the original
            return_value=constraints,
        ) as mock_get_outcome_constraint_transforms:
            gen_results = model.gen(
                n=1,
                search_space_digest=self.mf_search_space_digest,
                torch_opt_config=self.moo_torch_opt_config,
            )
        mock_get_outcome_constraint_transforms.assert_called_once()
        ckwargs = mock_get_outcome_constraint_transforms.call_args.kwargs
        oc = ckwargs["outcome_constraints"]
        self.assertTrue(torch.equal(oc[0], subset_outcome_constraints[0]))
        self.assertTrue(torch.equal(oc[1], subset_outcome_constraints[1]))

        # Check input constructor args
        ckwargs = mock_input_constructor.call_args.kwargs
        expected_kwargs = {
            "constraints",
            "bounds",
            "objective",
            "training_data",
            "X_baseline",
            "model",
            "objective_thresholds",
            "eta",
            "constraints_tuple",
        }
        self.assertSetEqual(set(ckwargs.keys()), expected_kwargs)
        for k in expected_kwargs:
            self.assertIsNotNone(ckwargs[k], f"{k} is None")

        self.assertIs(model.botorch_acqf_class, qLogNoisyExpectedHypervolumeImprovement)
        mock_input_constructor.assert_called_once()
        m = ckwargs["model"]
        self.assertIsInstance(m, SingleTaskGP)
        self.assertIsInstance(m.likelihood, FixedNoiseGaussianLikelihood)
        self.assertEqual(m.num_outputs, 2)
        training_data = ckwargs["training_data"]
        self.assertIsNotNone(training_data.Yvar)
        self.assertTrue(torch.equal(training_data.X, self.Xs))
        self.assertTrue(
            torch.equal(
                training_data.Y,
                torch.cat([ds.Y for ds in self.moo_training_data], dim=-1),
            )
        )
        self.assertTrue(
            torch.equal(
                training_data.Yvar,
                torch.cat([ds.Yvar for ds in self.moo_training_data], dim=-1),
            )
        )
        self.assertTrue(
            torch.equal(
                ckwargs["objective_thresholds"], self.moo_objective_thresholds[:2]
            )
        )
        self.assertIs(ckwargs["constraints"], constraints)

        obj_t = gen_results.gen_metadata["objective_thresholds"]
        self.assertTrue(torch.equal(obj_t[:2], self.moo_objective_thresholds[:2]))
        self.assertTrue(np.isnan(obj_t[2].item()))

        self.assertIsInstance(ckwargs["objective"], WeightedMCMultiOutputObjective)
        self.assertTrue(
            torch.equal(
                ckwargs["objective"].weights,
                self.moo_objective_weights[:2],
            )
        )
        expected_X_baseline = _filter_X_observed(
            Xs=[dataset.X for dataset in self.moo_training_data],
            objective_weights=self.moo_objective_weights,
            outcome_constraints=self.moo_outcome_constraints,
            bounds=self.search_space_digest.bounds,
            linear_constraints=self.linear_constraints,
            fixed_features=self.fixed_features,
        )
        self.assertTrue(
            torch.equal(ckwargs["X_baseline"], none_throws(expected_X_baseline))
        )
        # test inferred objective_thresholds
        objective_weights = torch.tensor([-1.0, -1.0, 0.0])
        outcome_constraints = (
            torch.tensor([[1.0, 0.0, 0.0]]),
            torch.tensor([[10.0]]),
        )
        linear_constraints = (
            torch.tensor([[1.0, 0.0, 0.0]]),
            torch.tensor([[2.0]]),
        )

        torch_opt_config = dataclasses.replace(
            self.moo_torch_opt_config,
            objective_weights=objective_weights,
            objective_thresholds=None,
            outcome_constraints=outcome_constraints,
            linear_constraints=linear_constraints,
        )

        objective_thresholds = torch.tensor([9.9, 3.3, float("nan")])
        with mock.patch(
            "ax.generators.torch.botorch_modular.acquisition"
            ".infer_objective_thresholds",
            return_value=objective_thresholds,
        ) as _mock_model_infer_objective_thresholds:
            gen_results = model.gen(
                n=1,
                search_space_digest=self.mf_search_space_digest,
                torch_opt_config=torch_opt_config,
            )
        expected_X_baseline = _filter_X_observed(
            Xs=[dataset.X for dataset in self.moo_training_data],
            objective_weights=objective_weights,
            outcome_constraints=outcome_constraints,
            bounds=self.search_space_digest.bounds,
            linear_constraints=linear_constraints,
            fixed_features=self.fixed_features,
        )
        ckwargs = _mock_model_infer_objective_thresholds.call_args[1]
        self.assertTrue(
            torch.equal(
                ckwargs["objective_weights"],
                objective_weights,
            )
        )
        oc = ckwargs["outcome_constraints"]
        self.assertTrue(torch.equal(oc[0], outcome_constraints[0]))
        self.assertTrue(torch.equal(oc[1], outcome_constraints[1]))
        m = ckwargs["model"]
        self.assertIsInstance(m, SingleTaskGP)
        self.assertIsInstance(m.likelihood, FixedNoiseGaussianLikelihood)
        self.assertEqual(m.num_outputs, 2)
        self.assertIn("objective_thresholds", gen_results.gen_metadata)
        obj_t = gen_results.gen_metadata["objective_thresholds"]
        self.assertTrue(torch.equal(obj_t[:2], objective_thresholds[:2]))
        self.assertTrue(np.isnan(obj_t[2].item()))

        # Avoid polluting the registry for other tests; re-register correct input
        # constructor for qLogNEHVI.
        _register_acqf_input_constructor(
            acqf_cls=qLogNoisyExpectedHypervolumeImprovement,
            input_constructor=qLogNEHVI_input_constructor,
        )

    @skip_if_import_error
    @mock_botorch_optimize
    def test_gen_multi_acquisition(self) -> None:
        # import pymoo here to raise an import error if pymoo is not installed.
        # The import error is already handled in
        # ax/generators/torch/botorch_modular/acquisition.py so we need to
        # reimport it here
        import pymoo  # noqa: F401

        botorch_acqf_classes_with_options = [
            (PosteriorMean, {}),
            (qLogNoisyExpectedImprovement, self.botorch_acqf_options),
        ]
        surrogate = Surrogate()
        model = BoTorchGenerator(
            surrogate=surrogate,
            acquisition_class=MultiAcquisition,
            botorch_acqf_classes_with_options=botorch_acqf_classes_with_options,
        )

        qLogNEI_input_constructor = get_acqf_input_constructor(
            qLogNoisyExpectedImprovement
        )
        mock_input_constructor = mock.MagicMock(
            qLogNEI_input_constructor, side_effect=qLogNEI_input_constructor
        )
        _register_acqf_input_constructor(
            acqf_cls=qLogNoisyExpectedImprovement,
            input_constructor=mock_input_constructor,
        )

        pm_input_constructor = get_acqf_input_constructor(PosteriorMean)
        mock_pm_input_constructor = mock.MagicMock(
            pm_input_constructor, side_effect=pm_input_constructor
        )
        _register_acqf_input_constructor(
            acqf_cls=PosteriorMean,
            input_constructor=mock_pm_input_constructor,
        )
        # Assert that error is raised if we haven't fit the model
        with self.assertRaisesRegex(ModelError, "fit the model first"):
            model.gen(
                n=1,
                search_space_digest=self.search_space_digest,
                torch_opt_config=self.torch_opt_config,
            )
        model.fit(
            datasets=self.block_design_training_data,
            search_space_digest=self.search_space_digest,
        )
        with ExitStack() as es:
            mock_init_acqf = es.enter_context(
                mock.patch.object(
                    BoTorchGenerator,
                    "_instantiate_acquisition",
                    wraps=model._instantiate_acquisition,
                )
            )

            gen_results = model.gen(
                n=1,
                search_space_digest=self.search_space_digest,
                torch_opt_config=self.torch_opt_config,
            )
            self.assertTrue(len(gen_results.points), 1)
            # check that we return both AF values for each point
            self.assertTrue(
                len(gen_results.gen_metadata["expected_acquisition_value"]), 1
            )
            self.assertTrue(
                len(gen_results.gen_metadata["expected_acquisition_value"][0]), 2
            )
        # Assert qLogNEI is initialized with expected arguments
        mock_init_acqf.assert_called_once_with(
            search_space_digest=self.search_space_digest,
            torch_opt_config=self.torch_opt_config,
            botorch_acqf_options={},
            acq_options={Keys.SUBSET_MODEL: False},
            botorch_acqf_classes_with_options=botorch_acqf_classes_with_options,
            n=1,
        )

        mock_input_constructor.assert_called_once()
        ckwargs = mock_input_constructor.call_args[1]

        # We particularly want to make sure that args that will not be used are
        # not passed
        expected_kwargs = {
            "bounds",
            "constraints",
            "X_baseline",
            "sampler",
            "objective",
            "training_data",
            "model",
            "constraints_tuple",
        }
        self.assertSetEqual(set(ckwargs.keys()), expected_kwargs)
        for k in expected_kwargs:
            self.assertIsNotNone(ckwargs[k], f"{k} is None")

        m = ckwargs["model"]
        self.assertIsInstance(m, SingleTaskGP)
        self.assertEqual(m.num_outputs, 1)
        training_data = ckwargs["training_data"]
        self.assertIsInstance(training_data, SupervisedDataset)
        self.assertTrue(torch.equal(training_data.X, self.Xs))
        self.assertTrue(
            torch.equal(
                training_data.Y,
                torch.cat([ds.Y for ds in self.block_design_training_data], dim=-1),
            )
        )

        self.assertIsInstance(ckwargs["objective"], GenericMCObjective)
        expected_X_baseline = _filter_X_observed(
            Xs=[dataset.X for dataset in self.block_design_training_data],
            objective_weights=self.objective_weights,
            outcome_constraints=self.outcome_constraints,
            bounds=self.search_space_digest.bounds,
            linear_constraints=self.linear_constraints,
            fixed_features=self.fixed_features,
        )
        self.assertTrue(
            torch.equal(ckwargs["X_baseline"], none_throws(expected_X_baseline))
        )
        # Assert PosteriorMean is initialized with expected arguments
        mock_pm_input_constructor.assert_called_once()
        ckwargs = mock_pm_input_constructor.call_args[1]
        # We particularly want to make sure that args that will not be used are
        # not passed
        expected_kwargs = {
            "bounds",
            "X_baseline",
            "training_data",
            "model",
            "posterior_transform",
            "constraints",
            "constraints_tuple",
        }
        self.assertSetEqual(set(ckwargs.keys()), expected_kwargs)
        for k in expected_kwargs:
            self.assertIsNotNone(ckwargs[k], f"{k} is None")

        m = ckwargs["model"]
        self.assertIsInstance(m, SingleTaskGP)
        self.assertEqual(m.num_outputs, 1)
        training_data = ckwargs["training_data"]
        self.assertIsInstance(training_data, SupervisedDataset)
        self.assertTrue(torch.equal(training_data.X, self.Xs))
        self.assertTrue(
            torch.equal(
                training_data.Y,
                torch.cat([ds.Y for ds in self.block_design_training_data], dim=-1),
            )
        )

        self.assertIsInstance(
            ckwargs["posterior_transform"], ScalarizedPosteriorTransform
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from abc import ABCMeta, abstractmethod, abstractproperty
from copy import deepcopy
from enum import Enum
from logging import Logger
from math import inf
from typing import cast, Union
from warnings import warn

from ax.core.types import TNumeric, TParameterization, TParamValue, TParamValueList
from ax.exceptions.core import AxParameterWarning, UnsupportedError, UserInputError
from ax.utils.common.base import SortableBase
from ax.utils.common.logger import get_logger

from ax.utils.common.string_utils import sanitize_name, unsanitize_name
from pyre_extensions import assert_is_instance, none_throws
from sympy.core.add import Add
from sympy.core.mul import Mul
from sympy.core.numbers import Float, Integer
from sympy.core.symbol import Symbol
from sympy.core.sympify import sympify

logger: Logger = get_logger(__name__)

# Tolerance for floating point comparisons. This is relatively permissive,
# and allows for serializing at rather low numerical precision.
# TODO: Do a more comprehensive audit of how floating point precision issues
# may creep up and implement a more principled fix
EPS = 1.5e-7
MAX_VALUES_CHOICE_PARAM = 1000
FIXED_CHOICE_PARAM_ERROR = (
    "ChoiceParameters require multiple feasible values. "
    "Please use FixedParameter instead when setting a single possible value."
)
REPR_FLAGS_IF_TRUE_ONLY = [
    "is_fidelity",
    "is_task",
    "is_hierarchical",
    "log_scale",
    "logit_scale",
]


class ParameterType(Enum):
    # pyre-fixme[35]: Target cannot be annotated.
    BOOL: int = 0
    # pyre-fixme[35]: Target cannot be annotated.
    INT: int = 1
    # pyre-fixme[35]: Target cannot be annotated.
    FLOAT: int = 2
    # pyre-fixme[35]: Target cannot be annotated.
    STRING: int = 3

    @property
    def is_numeric(self) -> bool:
        return self == ParameterType.INT or self == ParameterType.FLOAT


TParameterType = Union[type[int], type[float], type[str], type[bool]]

# pyre: PARAMETER_PYTHON_TYPE_MAP is declared to have type
# pyre: `Dict[ParameterType, Union[Type[bool], Type[float], Type[int],
# pyre: Type[str]]]` but is used as type `Dict[ParameterType,
# pyre-fixme[9]: Type[Union[float, str]]]`.
PARAMETER_PYTHON_TYPE_MAP: dict[ParameterType, TParameterType] = {
    ParameterType.INT: int,
    ParameterType.FLOAT: float,
    ParameterType.STRING: str,
    ParameterType.BOOL: bool,
}

SUPPORTED_PARAMETER_TYPES: tuple[
    type[bool] | type[float] | type[int] | type[str], ...
] = tuple(PARAMETER_PYTHON_TYPE_MAP.values())


# pyre-fixme[24]: Generic type `type` expects 1 type parameter, use `typing.Type` to
#  avoid runtime subscripting errors.
def _get_parameter_type(python_type: type) -> ParameterType:
    """Given a Python type, retrieve corresponding Ax ``ParameterType``."""
    for param_type, py_type in PARAMETER_PYTHON_TYPE_MAP.items():
        if py_type == python_type:
            return param_type
    raise ValueError(f"No Ax parameter type corresponding to {python_type}.")


class Parameter(SortableBase, metaclass=ABCMeta):
    _is_fidelity: bool = False
    _name: str
    _target_value: TParamValue = None
    _parameter_type: ParameterType

    def cast(self, value: TParamValue) -> TParamValue:
        if value is None:
            raise UnsupportedError("None values are not supported.")
        return self.python_type(value)

    @abstractmethod
    def validate(self, value: TParamValue, raises: bool = False) -> bool:
        """Returns True if input is a valid value for the parameter.

        Args:
            value: Value being checked.
            raises: If true, and validation fails, raises a UserInputError.

        Raises:
            UserInputError: If validation fails and raises is True.

        Returns:
            True if valid, False otherwise.
        """

    @abstractmethod
    def cardinality(self) -> float:
        pass

    @property
    def python_type(self) -> TParameterType:
        """The python type for the corresponding ParameterType enum.

        Used primarily for casting values of unknown type to conform
        to that of the parameter.
        """
        return PARAMETER_PYTHON_TYPE_MAP[self.parameter_type]

    def is_valid_type(self, value: TParamValue) -> bool:
        """Whether a given value's type is allowed by this parameter."""
        return type(value) is self.python_type

    @property
    def is_numeric(self) -> bool:
        return self.parameter_type.is_numeric

    @property
    def is_fidelity(self) -> bool:
        return self._is_fidelity

    @property
    def is_hierarchical(self) -> bool:
        return isinstance(self, (ChoiceParameter, FixedParameter)) and bool(
            self._dependents
        )

    @property
    def target_value(self) -> TParamValue:
        return self._target_value

    @property
    def parameter_type(self) -> ParameterType:
        return self._parameter_type

    @property
    def name(self) -> str:
        return self._name

    @property
    def dependents(self) -> dict[TParamValue, list[str]]:
        raise NotImplementedError(
            "Only choice hierarchical parameters are currently supported."
        )

    # pyre-fixme[7]: Expected `Parameter` but got implicit return value of `None`.
    def clone(self) -> Parameter:
        pass

    @property
    def _unique_id(self) -> str:
        return str(self)

    def _base_repr(self) -> str:
        ret_val = (
            f"{self.__class__.__name__}("
            f"name='{self._name}', "
            f"parameter_type={self.parameter_type.name}, "
            f"{self.domain_repr}"
        )

        # Add binary flags.
        for flag in self.available_flags:
            val = getattr(self, flag, False)
            if flag not in REPR_FLAGS_IF_TRUE_ONLY or (
                flag in REPR_FLAGS_IF_TRUE_ONLY and val is True
            ):
                ret_val += f", {flag}={val}"

        # Add target_value if one exists.
        if self.target_value is not None:
            tval_rep = self.target_value
            if self.parameter_type == ParameterType.STRING:
                tval_rep = f"'{tval_rep}'"
            ret_val += f", target_value={tval_rep}"

        return ret_val

    @abstractproperty
    def domain_repr(self) -> str:
        """Returns a string representation of the domain."""
        pass

    @property
    def available_flags(self) -> list[str]:
        """List of boolean attributes that can be set on this parameter."""
        return ["is_fidelity"]

    @property
    def summary_dict(
        self,
    ) -> dict[str, TParamValueList | TParamValue | str | list[str]]:
        # Assemble dict.
        summary_dict = {
            "name": self.name,
            "type": self.__class__.__name__.removesuffix("Parameter"),
            "domain": self.domain_repr,
            "parameter_type": self.parameter_type.name.lower(),
        }

        # Extract flags.
        flags = []
        for flag in self.available_flags:
            flag_val = getattr(self, flag, None)
            flag_repr = flag.removeprefix("is_")
            if flag == "sort_values":
                flag_repr = "sorted"
            if flag_val is True:
                flags.append(flag_repr)
            elif flag_val is False and flag not in REPR_FLAGS_IF_TRUE_ONLY:
                flags.append("un" + flag_repr)

        # Add flags, target_values, and dependents if present.
        if flags:
            summary_dict["flags"] = ", ".join(flags)
        if getattr(self, "is_fidelity", False) or getattr(self, "is_task", False):
            # pyre-fixme[6]: For 2nd argument expected `str` but got `Union[None,
            #  bool, float, int, str]`.
            summary_dict["target_value"] = self.target_value
        if getattr(self, "is_hierarchical", False):
            # pyre-fixme[6]: For 2nd argument expected `str` but got
            #  `Dict[Union[None, bool, float, int, str], List[str]]`.
            summary_dict["dependents"] = self.dependents

        # pyre-fixme[7]: Expected `Dict[str, Union[None, List[Union[None, bool,
        #  float, int, str]], List[str], bool, float, int, str]]` but got `Dict[str,
        #  str]`.
        return summary_dict


class RangeParameter(Parameter):
    """Parameter object that specifies a range of values."""

    def __init__(
        self,
        name: str,
        parameter_type: ParameterType,
        lower: float,
        upper: float,
        log_scale: bool = False,
        logit_scale: bool = False,
        digits: int | None = None,
        is_fidelity: bool = False,
        target_value: TParamValue = None,
    ) -> None:
        """Initialize RangeParameter

        Args:
            name: Name of the parameter.
            parameter_type: Enum indicating the type of parameter
                value (e.g. string, int).
            lower: Lower bound of the parameter range (inclusive).
            upper: Upper bound of the parameter range (inclusive).
            log_scale: Whether to sample in the log space when drawing
                random values of the parameter.
            logit_scale: Whether to sample in logit space when drawing
                random values of the parameter.
            digits: Number of digits to round values to for float type.
            is_fidelity: Whether this parameter is a fidelity parameter.
            target_value: Target value of this parameter if it is a fidelity.
        """
        if is_fidelity and (target_value is None):
            raise UserInputError(
                "`target_value` should not be None for the fidelity parameter: "
                "{}".format(name)
            )

        self._name = name
        if parameter_type not in (ParameterType.INT, ParameterType.FLOAT):
            raise UserInputError("RangeParameter type must be int or float.")
        self._parameter_type = parameter_type
        self._digits = digits
        self._lower: TNumeric = self.cast(lower)
        self._upper: TNumeric = self.cast(upper)
        self._log_scale = log_scale
        self._logit_scale = logit_scale
        self._is_fidelity = is_fidelity
        self._target_value: TNumeric | None = (
            self.cast(target_value) if target_value is not None else None
        )

        self._validate_range_param(
            parameter_type=parameter_type,
            lower=lower,
            upper=upper,
            log_scale=log_scale,
            logit_scale=logit_scale,
        )

    def cardinality(self) -> TNumeric:
        if self.parameter_type == ParameterType.FLOAT:
            return inf

        return int(self.upper) - int(self.lower) + 1

    def _validate_range_param(
        self,
        lower: TNumeric,
        upper: TNumeric,
        log_scale: bool,
        logit_scale: bool,
        parameter_type: ParameterType | None = None,
    ) -> None:
        if parameter_type and parameter_type not in (
            ParameterType.INT,
            ParameterType.FLOAT,
        ):
            raise UserInputError(
                f"RangeParameter {self.name}type must be int or float."
            )

        upper = float(upper)
        if lower >= upper:
            raise UserInputError(
                f"Upper bound of {self.name} must be strictly larger than lower."
                f"Got: ({lower}, {upper})."
            )
        width: float = upper - lower
        if width < 100 * EPS:
            raise UserInputError(
                f"Parameter {self.name}'s range ({width}) is very small and likely "
                "to cause numerical errors. Consider reparameterizing your "
                "problem by scaling the parameter."
            )
        if log_scale and logit_scale:
            raise UserInputError(f"{self.name} can't use both log and logit.")
        if log_scale and lower <= 0:
            raise UserInputError(f"{self.name} cannot take log when min <= 0.")
        if logit_scale and (lower <= 0 or upper >= 1):
            raise UserInputError(f"{self.name} logit requires lower > 0 and upper < 1")
        if not (self.is_valid_type(lower)) or not (self.is_valid_type(upper)):
            raise UserInputError(
                f"[{lower}, {upper}] is an invalid range for {self.name}."
            )

    @property
    def upper(self) -> TNumeric:
        """Upper bound of the parameter range.

        Value is cast to parameter type upon set and also validated
        to ensure the bound is strictly greater than lower bound.
        """
        return self._upper

    @upper.setter
    def upper(self, value: TNumeric) -> None:
        self._validate_range_param(
            lower=self.lower,
            upper=value,
            log_scale=self.log_scale,
            logit_scale=self.logit_scale,
        )
        self._upper = self.cast(value)

    @property
    def lower(self) -> TNumeric:
        """Lower bound of the parameter range.

        Value is cast to parameter type upon set and also validated
        to ensure the bound is strictly less than upper bound.
        """
        return self._lower

    @lower.setter
    def lower(self, value: TNumeric) -> None:
        self._validate_range_param(
            lower=value,
            upper=self.upper,
            log_scale=self.log_scale,
            logit_scale=self.logit_scale,
        )
        self._lower = self.cast(value)

    @property
    def digits(self) -> int | None:
        """Number of digits to round values to for float type.

        Upper and lower bound are re-cast after this property is changed.
        """
        return self._digits

    @property
    def log_scale(self) -> bool:
        """Whether the parameter's random values should be sampled from log space."""
        return self._log_scale

    @property
    def logit_scale(self) -> bool:
        """Whether the parameter's random values should be sampled from logit space."""
        return self._logit_scale

    def update_range(
        self, lower: float | None = None, upper: float | None = None
    ) -> RangeParameter:
        """Set the range to the given values.

        If lower or upper is not provided, it will be left at its current value.

        Args:
            lower: New value for the lower bound.
            upper: New value for the upper bound.
        """
        if lower is None:
            lower = self._lower
        if upper is None:
            upper = self._upper

        cast_lower = self.cast(lower)
        cast_upper = self.cast(upper)
        self._validate_range_param(
            lower=cast_lower,
            upper=cast_upper,
            log_scale=self.log_scale,
            logit_scale=self.logit_scale,
        )
        self._lower = cast_lower
        self._upper = cast_upper
        return self

    def set_digits(self, digits: int | None) -> RangeParameter:
        self._digits = digits

        # Re-scale min and max to new digits definition
        cast_lower = self.cast(self._lower)
        cast_upper = self.cast(self._upper)
        # `<=` is not supported for operand types `Union[float, int]` and `int`.
        # pyre-ignore [58]
        if cast_lower >= cast_upper:
            raise UserInputError(
                f"Lower bound {cast_lower} is >= upper bound {cast_upper}."
            )

        self._lower = cast_lower
        self._upper = cast_upper
        return self

    def set_log_scale(self, log_scale: bool) -> RangeParameter:
        self._log_scale = log_scale
        return self

    def set_logit_scale(self, logit_scale: bool) -> RangeParameter:
        self._logit_scale = logit_scale
        return self

    def validate(
        self, value: TParamValue, raises: bool = False, tol: float = EPS
    ) -> bool:
        """Returns True if input is a valid value for the parameter.

        Checks that value is of the right type and within
        the valid range for the parameter. Returns False if value is None.

        Args:
            value: Value being checked.
            raises: If true, and validation fails, raises a UserInputError.
            tol: Absolute tolerance for floating point comparisons.

        Raises:
            UserInputError: If validation fails and raises is True.

        Returns:
            True if valid, False otherwise.
        """

        def return_false_or_raise(msg: str) -> bool:
            if raises:
                raise UserInputError(msg)
            return False

        if value is None:
            msg = (
                f"Value of parameter {self.name} is `None` but the parameter "
                f"type is {self.parameter_type}."
            )
            return return_false_or_raise(msg)

        if not self.is_valid_type(value):
            msg = (
                f"Value ({value}) of parameter {self.name} has type ({type(value)}), "
                "which is not valid for a RangeParameter with parameter type "
                f"{self.parameter_type}."
            )
            return return_false_or_raise(msg)

        value = self.cast(value)
        if value < self.lower - tol or value > self.upper + tol:
            interval = (self.lower, self.upper)
            msg = (
                f"Value ({value}) of parameter {self.name} is not within the range of "
                f"the parameter {interval}, even with a tolerance of {tol}."
            )
            return return_false_or_raise(msg)

        return True

    def is_valid_type(self, value: TParamValue) -> bool:
        """Same as default except allows floats whose value is an int
        for Int parameters.
        """
        if not (isinstance(value, float) or isinstance(value, int)):
            return False

        # This might have issues with ints > 2^24
        if self.parameter_type is ParameterType.INT:
            return isinstance(value, int) or float(none_throws(value)).is_integer()
        return True

    def clone(self) -> RangeParameter:
        return RangeParameter(
            name=self._name,
            parameter_type=self._parameter_type,
            lower=self._lower,
            upper=self._upper,
            log_scale=self._log_scale,
            logit_scale=self._logit_scale,
            digits=self._digits,
            is_fidelity=self._is_fidelity,
            target_value=self._target_value,
        )

    def cast(self, value: TParamValue) -> TNumeric:
        value = super().cast(value=value)
        if self.parameter_type is ParameterType.FLOAT and self._digits is not None:
            return round(float(value), none_throws(self._digits))
        return assert_is_instance(value, TNumeric)

    def __repr__(self) -> str:
        ret_val = self._base_repr()

        if self._digits is not None:
            ret_val += f", digits={self._digits}"

        return ret_val + ")"

    @property
    def available_flags(self) -> list[str]:
        """List of boolean attributes that can be set on this parameter."""
        return super().available_flags + ["log_scale", "logit_scale"]

    @property
    def domain_repr(self) -> str:
        """Returns a string representation of the domain."""
        return f"range={[self.lower, self.upper]}"


class ChoiceParameter(Parameter):
    """Parameter object that specifies a discrete set of values.

    Args:
        name: Name of the parameter.
        parameter_type: Enum indicating the type of parameter
            value (e.g. string, int).
        values: List of allowed values for the parameter.
        is_ordered: If False, the parameter is a categorical variable.
            Defaults to False if parameter_type is STRING and ``values``
            is longer than 2, else True.
        is_task: Treat the parameter as a task parameter for modeling.
        is_fidelity: Whether this parameter is a fidelity parameter.
        target_value: Target value of this parameter if it's a fidelity or
            task parameter.
        sort_values: Whether to sort ``values`` before encoding.
            Defaults to False if ``parameter_type`` is STRING, else
            True.
        dependents: Optional mapping for parameters in hierarchical search
            spaces; format is { value -> list of dependent parameter names }.
    """

    def __init__(
        self,
        name: str,
        parameter_type: ParameterType,
        values: list[TParamValue],
        is_ordered: bool | None = None,
        is_task: bool = False,
        is_fidelity: bool = False,
        target_value: TParamValue = None,
        sort_values: bool | None = None,
        dependents: dict[TParamValue, list[str]] | None = None,
    ) -> None:
        if (is_fidelity or is_task) and (target_value is None):
            ptype = "fidelity" if is_fidelity else "task"
            raise UserInputError(
                f"`target_value` should not be None for the {ptype} parameter: "
                "{}".format(name)
            )

        self._name = name
        self._parameter_type = parameter_type
        self._is_task = is_task
        self._is_fidelity = is_fidelity
        self._target_value: TParamValue = (
            self.cast(target_value) if target_value is not None else None
        )
        # A choice parameter with only one value is a FixedParameter.
        if not len(values) > 1:
            raise UserInputError(f"{self._name}({values}): {FIXED_CHOICE_PARAM_ERROR}")
        # Cap the number of possible values.
        if len(values) > MAX_VALUES_CHOICE_PARAM:
            raise UserInputError(
                f"`ChoiceParameter` with more than {MAX_VALUES_CHOICE_PARAM} values "
                "is not supported! Use a `RangeParameter` instead."
            )
        # Remove duplicate values.
        # Using dict to deduplicate here since set doesn't preserve order but dict does.
        dict_values = dict.fromkeys(values)
        if len(values) != len(dict_values):
            warn(
                f"Duplicate values found for ChoiceParameter {name}. "
                "Initializing the parameter with duplicate values removed. ",
                AxParameterWarning,
                stacklevel=2,
            )
            values = list(dict_values)

        if is_ordered is False and len(values) == 2:
            is_ordered = True
            logger.debug(
                f"Changing `is_ordered` to `True` for `ChoiceParameter` '{name}' since "
                "there are only two possible values.",
                AxParameterWarning,
                stacklevel=3,
            )
        self._is_ordered: bool = (
            is_ordered
            if is_ordered is not None
            else self._get_default_is_ordered_and_warn(num_choices=len(values))
        )
        # sort_values defaults to True if the parameter is not a string
        self._sort_values: bool = (
            sort_values
            if sort_values is not None
            else self._get_default_sort_values_and_warn()
        )
        if self.sort_values:
            values = cast(list[TParamValue], sorted([none_throws(v) for v in values]))
        self._values: list[TParamValue] = self._cast_values(values)

        if dependents:
            for value in dependents:
                if value not in self.values:
                    raise UserInputError(
                        f"Value {value} in `dependents` "
                        f"argument is not among the parameter values: {self.values}."
                    )
        # NOTE: We don't need to check that dependent parameters actually exist as
        # that is done in `HierarchicalSearchSpace` constructor.
        self._dependents = dependents

    def _get_default_is_ordered_and_warn(self, num_choices: int) -> bool:
        default_bool = self._parameter_type != ParameterType.STRING or num_choices == 2
        if self._parameter_type == ParameterType.STRING and num_choices > 2:
            motivation = " since the parameter is a string with more than 2 choices."
        elif num_choices == 2:
            motivation = " since there are exactly two choices."
        else:
            motivation = " since the parameter is not of type string."
        warn(
            f'`is_ordered` is not specified for `ChoiceParameter` "{self._name}". '
            f"Defaulting to `{default_bool}` {motivation}. To override this behavior "
            f"(or avoid this warning), specify `is_ordered` during `ChoiceParameter` "
            "construction. Note that choice parameters with exactly 2 choices are "
            "always considered ordered and that the user-supplied `is_ordered` has no "
            "effect in this particular case.",
            AxParameterWarning,
            stacklevel=3,
        )
        return default_bool

    def _get_default_sort_values_and_warn(self) -> bool:
        default_bool = self._parameter_type != ParameterType.STRING
        warn(
            f'`sort_values` is not specified for `ChoiceParameter` "{self._name}". '
            f"Defaulting to `{default_bool}` for parameters of `ParameterType` "
            f"{self.parameter_type.name}. To override this behavior (or avoid this "
            f"warning), specify `sort_values` during `ChoiceParameter` construction.",
            AxParameterWarning,
            stacklevel=3,
        )
        return default_bool

    def cardinality(self) -> float:
        return len(self.values)

    @property
    def sort_values(self) -> bool:
        return self._sort_values

    @property
    def is_ordered(self) -> bool:
        return self._is_ordered

    @property
    def is_task(self) -> bool:
        return self._is_task

    @property
    def values(self) -> list[TParamValue]:
        return self._values

    def set_values(self, values: list[TParamValue]) -> ChoiceParameter:
        """Set the list of allowed values for parameter.

        Cast all input values to the parameter type.

        Args:
            values: New list of allowed values.
        """
        # A choice parameter with only one value is a FixedParameter.
        if not len(values) > 1:
            raise UserInputError(FIXED_CHOICE_PARAM_ERROR)
        self._values = self._cast_values(values)
        return self

    def add_values(self, values: list[TParamValue]) -> ChoiceParameter:
        """Add input list to the set of allowed values for parameter.

        Cast all input values to the parameter type.

        Args:
            values: Values being added to the allowed list.
        """
        self._values.extend(self._cast_values(values))
        return self

    def validate(self, value: TParamValue, raises: bool = False) -> bool:
        """Checks that the input is in the list of allowed values.

        Args:
            value: Value being checked.
            raises: If true, and validation fails, raises a UserInputError.

        Raises:
            UserInputError: If validation fails and raises is True.

        Returns:
            True if valid, False otherwise.
        """
        is_valid = value in self._values
        if raises and not is_valid:
            raise UserInputError(
                f"Value {value} is not in the list of allowed values: {self._values}."
            )
        return is_valid

    @property
    def dependents(self) -> dict[TParamValue, list[str]]:
        if not self.is_hierarchical:
            raise NotImplementedError(
                "Only hierarchical parameters support the `dependents` property."
            )
        return none_throws(self._dependents)

    def _cast_values(self, values: list[TParamValue]) -> list[TParamValue]:
        return [self.cast(value) for value in values]

    def clone(self) -> ChoiceParameter:
        return ChoiceParameter(
            name=self._name,
            parameter_type=self._parameter_type,
            values=self._values,
            is_ordered=self._is_ordered,
            is_task=self._is_task,
            is_fidelity=self._is_fidelity,
            target_value=self._target_value,
            sort_values=self._sort_values,
            dependents=deepcopy(self._dependents),
        )

    def __repr__(self) -> str:
        ret_val = self._base_repr()

        if self._dependents:
            ret_val += f", dependents={self._dependents}"

        return ret_val + ")"

    @property
    def available_flags(self) -> list[str]:
        """List of boolean attributes that can be set on this parameter."""
        return super().available_flags + [
            "is_ordered",
            "is_hierarchical",
            "is_task",
            "sort_values",
        ]

    @property
    def domain_repr(self) -> str:
        """Returns a string representation of the domain."""
        return f"values={self.values}"


class FixedParameter(Parameter):
    """Parameter object that specifies a single fixed value."""

    def __init__(
        self,
        name: str,
        parameter_type: ParameterType,
        value: TParamValue,
        is_fidelity: bool = False,
        target_value: TParamValue = None,
        dependents: dict[TParamValue, list[str]] | None = None,
    ) -> None:
        """Initialize FixedParameter

        Args:
            name: Name of the parameter.
            parameter_type: Enum indicating the type of parameter
                value (e.g. string, int).
            value: The fixed value of the parameter.
            is_fidelity: Whether this parameter is a fidelity parameter.
            target_value: Target value of this parameter if it is a fidelity.
            dependents: Optional mapping for parameters in hierarchical search
                spaces; format is { value -> list of dependent parameter names }.
        """
        if is_fidelity and (target_value is None):
            raise UserInputError(
                "`target_value` should not be None for the fidelity parameter: "
                "{}".format(name)
            )

        self._name = name
        self._parameter_type = parameter_type
        self._value: TParamValue = self.cast(value)
        self._is_fidelity = is_fidelity
        self._target_value: TParamValue = (
            self.cast(target_value) if target_value is not None else None
        )
        # NOTE: We don't need to check that dependent parameters actually exist as
        # that is done in `HierarchicalSearchSpace` constructor.
        if dependents:
            if len(dependents) > 1 or next(iter(dependents.keys())) != self.value:
                raise UserInputError(
                    "The only expected key in `dependents` for fixed parameter "
                    f"{self.name}: {self.value}; got: {dependents}."
                )
        self._dependents = dependents

    def cardinality(self) -> float:
        return 1.0

    @property
    def value(self) -> TParamValue:
        return self._value

    def set_value(self, value: TParamValue) -> FixedParameter:
        self._value = self.cast(value)
        return self

    def validate(self, value: TParamValue, raises: bool = False) -> bool:
        """Checks that the input is equal to the fixed value.

        Args:
            value: Value being checked.
            raises: If true, and validation fails, raises a UserInputError.

        Raises:
            UserInputError: If validation fails and raises is True.

        Returns:
            True if valid, False otherwise.
        """
        is_valid = value == self._value
        if raises and not is_valid:
            raise UserInputError(
                f"Value {value} is not equal to the fixed value: {self._value}."
            )
        return is_valid

    @property
    def dependents(self) -> dict[TParamValue, list[str]]:
        if not self.is_hierarchical:
            raise NotImplementedError(
                "Only hierarchical parameters support the `dependents` property."
            )
        return none_throws(self._dependents)

    def clone(self) -> FixedParameter:
        return FixedParameter(
            name=self._name,
            parameter_type=self._parameter_type,
            value=self._value,
            is_fidelity=self._is_fidelity,
            target_value=self._target_value,
            dependents=self._dependents,
        )

    def __repr__(self) -> str:
        ret_val = self._base_repr()
        return ret_val + ")"

    @property
    def available_flags(self) -> list[str]:
        """List of boolean attributes that can be set on this parameter."""
        return super().available_flags + ["is_hierarchical"]

    @property
    def domain_repr(self) -> str:
        """Returns a string representation of the domain."""
        if self._parameter_type == ParameterType.STRING:
            return f"value='{self._value}'"
        else:
            return f"value={self._value}"


class DerivedParameter(Parameter):
    """A parameter that is a deterministic function of other parameters.

    Currently, only support for linear functions are implemented, but this is
    extendable to non-linear functions.
    """

    # pyre-fixme [13]: Uninitialized attribute [13]: Attribute `_intercept` is
    # declared in class `DerivedParameter` to have type `float` but is never
    # initialized.
    _intercept: float
    # pyre-fixme [13]: Uninitialized attribute [13]: Attribute
    # `_parameter_names_to_weights` is declared in class `DerivedParameter` to#
    # have type `typing.Dict[str, float]` but is never initialized.
    _parameter_names_to_weights: dict[str, float]

    def __init__(
        self,
        name: str,
        parameter_type: ParameterType,
        expression_str: str,
        is_fidelity: bool = False,
        target_value: TParamValue = None,
    ) -> None:
        """Initialize DerivedParameter

        Args:
            name: Name of the parameter.
            parameter_type: Enum indicating the type of parameter
                value (e.g. string, int).
            expression_str: A string expression of the derived parameter definition.
            is_fidelity: Whether this parameter is a fidelity parameter.
            target_value: Target value of this parameter if it is a fidelity.
        """
        if is_fidelity:
            raise UnsupportedError("Derived parameters cannot be fidelity parameters.")
        elif target_value is not None:
            raise UnsupportedError(
                "Derived parameters do not support specifying a target value."
            )
        elif parameter_type not in (ParameterType.FLOAT, ParameterType.INT):
            raise UserInputError(
                "Derived parameters must be of type float or int, but got "
                f"{parameter_type}."
            )

        self.set_expression_str(expression_str=expression_str)
        self._name = name
        self._parameter_type = parameter_type
        self._is_fidelity = is_fidelity
        self._target_value = target_value

    def _parse_expression_str(self, expression_str: str) -> None:
        """Parse the expression str into parameter names and coefficients.

        Currently only linear functions are supported.
        """
        expression = sympify(sanitize_name(expression_str))
        if isinstance(expression, (Float, Integer)):
            raise UserInputError(
                "Derived parameters must have at least one parameter in "
                "`expression_str`."
            )
        elif not isinstance(expression, (Add, Mul, Symbol)):
            raise UnsupportedError("Only linear expressions are currently supported.")
        coefficient_dict = expression.as_coefficients_dict()
        self._intercept = float(coefficient_dict.pop(1, 0.0))
        parameter_names_to_weights = {}
        for name, coef in coefficient_dict.items():
            if not isinstance(name, Symbol):
                raise UnsupportedError(
                    "Only linear expressions are currently supported."
                )
            parameter_names_to_weights[unsanitize_name(str(name))] = float(coef)
        self._parameter_names_to_weights = parameter_names_to_weights

    @property
    def domain_repr(self) -> str:
        """Returns a string representation of the derived parameter."""
        terms = [
            f"{weight} * {name}"
            for name, weight in self._parameter_names_to_weights.items()
        ]
        if self._intercept != 0.0:
            terms.append(str(self._intercept))
        return "value=" + " + ".join(terms)

    @property
    def parameter_names_to_weights(self) -> dict[str, float]:
        return self._parameter_names_to_weights

    @property
    def expression_str(self) -> str:
        return self._expression_str

    def set_expression_str(self, expression_str: str) -> None:
        self._expression_str = expression_str
        self._parse_expression_str(expression_str=expression_str)

    @property
    def intercept(self) -> float:
        return self._intercept

    def cardinality(self) -> float:
        if self.parameter_type == ParameterType.FLOAT:
            return inf
        raise UnsupportedError(
            "cardinality for an integer DerivedParameter is not supported."
        )

    def compute(self, parameters: TParameterization) -> TParamValue:
        """Compute the value of the derived parameter.

        Args:
            parameterization: A dictionary mapping parameter names to values.

        Returns:
            The value of the derived parameter.
        """
        return self.cast(
            self._intercept
            + sum(
                self._parameter_names_to_weights[parameter_name]
                * float(parameters[parameter_name])
                for parameter_name in self._parameter_names_to_weights
            )
        )

    def validate(
        self,
        value: TParamValue,
        raises: bool = False,
        parameters: TParameterization | None = None,
    ) -> bool:
        """Checks that the input is equal to the derived value.

        Args:
            value: Value being checked.
            raises: If true, and validation fails, raises a UserInputError.
            parameters: A dictionary mapping parameter names to values. Used to
                compute the expected derived parameter value.

        Raises:
            UserInputError: If validation fails and raises is True.

        Returns:
            True if valid, False otherwise.
        """
        is_valid = False
        if parameters is None:
            if raises:
                raise UserInputError(
                    "Must specify `parameters` to validate a derived parameter"
                )
            return False
        expected_value = self.compute(parameters=parameters)
        is_valid = expected_value == value
        if raises and not is_valid:
            raise UserInputError(
                f"Value {value} is not equal to the expected derived"
                f" value: {expected_value}."
            )
        return is_valid

    def clone(self) -> DerivedParameter:
        return DerivedParameter(
            name=self._name,
            parameter_type=self._parameter_type,
            expression_str=self._expression_str,
            is_fidelity=self._is_fidelity,
            target_value=self._target_value,
        )

    def __repr__(self) -> str:
        ret_val = self._base_repr()
        return ret_val + ")"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass
from typing import Literal

from ax.api.configs import ChoiceParameterConfig, RangeParameterConfig


@dataclass
class ExperimentStruct:
    """
    Allows specifying the search space of an experiment along
    with other metadata.
    """

    parameters: list[RangeParameterConfig | ChoiceParameterConfig]
    # Parameter constraints will be parsed via SymPy
    # Ex: "num_layers1 <= num_layers2", "compound_a + compound_b <= 1"
    parameter_constraints: list[str]

    name: str | None
    description: str | None
    experiment_type: str | None
    owner: str | None


@dataclass
class GenerationStrategyDispatchStruct:
    """
    Allows the user to configure the way candidates are generated during the experiment.
    This is used, along with the properties of the experiment, to determine the
    ``GenerationStrategy`` to use for candidate generation.

    Args:
        method: The generation method to use. Provides high level guidance to the
            underlying generation strategy dispatch logic, which is responsible for
            determinining the exact details. Available options are:
                - ``"fast"``, a faster generation method that uses the built-in
                    defaults from the Modular BoTorch Model without any model
                    selection.
                - ``"random_search"``, primarily intended for pure exploration
                    experiments, this method utilizes quasi-random Sobol sequences
                    for candidate generation.
        initialization_budget: The number of trials to use for initialization.
            If ``None``, a default budget of 5 trials is used.
        initialization_random_seed: The random seed to use with the Sobol generator
            that generates the initialization trials.
        initialize_with_center: If True, the center of the search space is used as the
            first point. The definition of center respects the scaling of the
            parameters. For discrete parameters, the median value is considered the
            center, with the later points being used to break ties.
        use_existing_trials_for_initialization: Whether to count all trials attached
            to the experiment as part of the initialization budget. For example,
            if 2 trials were manually attached to the experiment and this option
            is set to ``True``, we will only generate `initialization_budget - 2`
            additional trials for initialization.
        min_observed_initialization_trials: The minimum required number of
            initialization trials with observations before the generation strategy
            is allowed to transition away from the initialization phase.
            Defaults to `max(1, initialization_budget // 2)`.
        allow_exceeding_initialization_budget: This option determines the behavior
            of the generation strategy when the ``initialization_budget`` is exhausted
            and ``min_observed_initialization_trials`` is not met. If this is ``True``,
            the generation strategy will generate additional initialization trials when
            a new trial is requested, exceeding the specified ``initialization_budget``.
            If this is ``False``, the generation strategy will raise an error and the
            candidate generation may be continued when additional data is observed
            for the existing trials.
        torch_device: The device to use for model fitting and candidate
            generation in PyTorch / BoTorch based generation nodes.
            NOTE: This option is not validated. Please ensure that the string
            input corresponds to a valid device.
    """

    method: Literal["quality", "fast", "random_search"] = "fast"
    # Initialization options
    initialization_budget: int | None = None
    initialization_random_seed: int | None = None
    initialize_with_center: bool = True
    use_existing_trials_for_initialization: bool = True
    min_observed_initialization_trials: int | None = None
    allow_exceeding_initialization_budget: bool = False
    # Misc options
    torch_device: str | None = None



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Mapping
from typing import Protocol, runtime_checkable

from ax.core.types import TParamValue


@runtime_checkable
class TBenchmarkStepRuntimeFunction(Protocol):
    def __call__(self, params: Mapping[str, TParamValue]) -> float:
        """
        Return the runtime for each step.

        Each step within an arm will take the same amount of time.
        """
        ...



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
from ax.analysis.plotly.top_surfaces import TopSurfacesAnalysis
from ax.api.client import Client
from ax.api.configs import ChoiceParameterConfig, RangeParameterConfig
from ax.exceptions.core import UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_offline_experiments_subset,
    get_online_experiments_subset,
)
from ax.utils.testing.mock import mock_botorch_optimize
from ax.utils.testing.modeling_stubs import get_default_generation_strategy_at_MBM_node
from pyre_extensions import assert_is_instance, none_throws


class TestTopSurfacesAnalysis(TestCase):
    @mock_botorch_optimize
    def test_compute(self) -> None:
        client = Client()
        client.configure_experiment(
            name="foo",
            parameters=[
                RangeParameterConfig(
                    name="x1",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
                RangeParameterConfig(
                    name="x2",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
            ],
        )
        client.configure_optimization(objective="bar")

        for _ in range(6):
            for trial_index, parameterization in client.get_next_trials(
                max_trials=1
            ).items():
                client.complete_trial(
                    trial_index=trial_index,
                    raw_data={
                        "bar": assert_is_instance(parameterization["x1"], float)
                        - 2 * assert_is_instance(parameterization["x2"], float)
                    },
                )

        analysis = TopSurfacesAnalysis(metric_name="bar", order="first")

        with self.assertRaisesRegex(UserInputError, "requires an Experiment"):
            analysis.compute()

        with self.assertRaisesRegex(
            UserInputError, "Must provide either a GenerationStrategy or an Adapter"
        ):
            analysis.compute(experiment=client._experiment)

        cards = analysis.compute(
            experiment=client._experiment,
            generation_strategy=client._generation_strategy,
        ).flatten()

        # First card should be the sensitivity analysis.
        self.assertEqual(cards[0].title, "Sensitivity Analysis for bar")

        # Other cards should be slices.
        self.assertIn("vs. bar", cards[1].title)
        self.assertIn("vs. bar", cards[2].title)

        second = TopSurfacesAnalysis(metric_name="bar", order="second")

        second_card_group = second.compute(
            experiment=client._experiment,
            generation_strategy=client._generation_strategy,
        )

        self.assertEqual(
            {card.name for card in second_card_group.children},
            {
                "TopSurfaceAnalysisSlicePlots",
                "TopSurfaceAnalysisContourPlots",
                "SensitivityAnalysisPlot",
            },
        )

        with_surfaces = second_card_group.flatten()

        self.assertEqual(len(with_surfaces), 4)

        # First card should be the sensitivity analysis.
        self.assertEqual(with_surfaces[0].title, "Sensitivity Analysis for bar")

        # Other cards should be slices or contours.
        self.assertIn("vs. bar", with_surfaces[1].title)
        self.assertIn("vs. bar", with_surfaces[2].title)
        self.assertIn("vs. bar", with_surfaces[3].title)

    @mock_botorch_optimize
    @TestCase.ax_long_test(reason="Expensive to compute Sobol indicies")
    def test_compute_categorical_parameters(self) -> None:
        client = Client()
        client.configure_experiment(
            name="foo",
            parameters=[
                RangeParameterConfig(
                    name="x1",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
                ChoiceParameterConfig(
                    name="x2",
                    parameter_type="int",
                    values=[*range(10)],
                    is_ordered=False,
                ),
            ],
        )
        client.configure_optimization(objective="bar")

        for _ in range(6):
            for trial_index, parameterization in client.get_next_trials(
                max_trials=1
            ).items():
                client.complete_trial(
                    trial_index=trial_index,
                    raw_data={
                        "bar": assert_is_instance(parameterization["x1"], float)
                        - 2 * assert_is_instance(parameterization["x2"], int)
                    },
                )

        analysis = TopSurfacesAnalysis(metric_name="bar")

        cards = analysis.compute(
            experiment=client._experiment,
            generation_strategy=client._generation_strategy,
        ).flatten()

        # Only plot x1 vs bar since x2 is categorical.
        self.assertEqual(len(cards), 2)
        self.assertEqual(cards[0].title, "Sensitivity Analysis for bar")
        self.assertEqual(cards[1].title, "x1 vs. bar")

    @mock_botorch_optimize
    @TestCase.ax_long_test(reason="Expensive to compute Sobol indicies")
    def test_online(self) -> None:
        # Test TopSurfacesAnalysis can be computed for a variety of experiments
        # which resemble those we see in an online setting.
        order = "total"  # most common

        for experiment in get_online_experiments_subset():
            generation_strategy = get_default_generation_strategy_at_MBM_node(
                experiment=experiment
            )
            analysis = TopSurfacesAnalysis(
                # Select and arbitrary metric from the optimization config
                metric_name=none_throws(
                    experiment.optimization_config
                ).objective.metric_names[0],
                order=order,
                top_k=2,
            )

            _ = analysis.compute(
                experiment=experiment, generation_strategy=generation_strategy
            )

        # test with default top k
        for experiment in get_online_experiments_subset():
            generation_strategy = get_default_generation_strategy_at_MBM_node(
                experiment=experiment
            )
            analysis = TopSurfacesAnalysis(
                # Select and arbitrary metric from the optimization config
                metric_name=none_throws(
                    experiment.optimization_config
                ).objective.metric_names[0],
                order=order,
            )

            _ = analysis.compute(
                experiment=experiment, generation_strategy=generation_strategy
            )

    @mock_botorch_optimize
    @TestCase.ax_long_test(reason="Expensive to compute Sobol indicies")
    def test_offline(self) -> None:
        # Test TopSurfacesAnalysis can be computed for a variety of experiments
        # which resemble those we see in an offline setting, in analogous tests we
        # run all experiments with modifications to settings, however, this test
        # is slow and so we limit the number of permutations we validate.
        order = "total"  # most common

        for experiment in get_offline_experiments_subset():
            generation_strategy = get_default_generation_strategy_at_MBM_node(
                experiment=experiment
            )
            analysis = TopSurfacesAnalysis(
                # Select and arbitrary metric from the optimization config
                metric_name=none_throws(
                    experiment.optimization_config
                ).objective.metric_names[0],
                order=order,
                top_k=2,
            )

            _ = analysis.compute(
                experiment=experiment, generation_strategy=generation_strategy
            )

        # run the test with the default top k
        for experiment in get_offline_experiments_subset():
            generation_strategy = get_default_generation_strategy_at_MBM_node(
                experiment=experiment
            )
            analysis = TopSurfacesAnalysis(
                # Select and arbitrary metric from the optimization config
                metric_name=none_throws(
                    experiment.optimization_config
                ).objective.metric_names[0],
                order=order,
            )

            _ = analysis.compute(
                experiment=experiment, generation_strategy=generation_strategy
            )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.utils.common.func_enum import FuncEnum
from ax.utils.common.testutils import TestCase


def morph_into_salamander(how_soon: int) -> bool:
    if how_soon <= 0:  # Now is the time!
        return True
    else:  # Not morphing yet
        return False


class AnimalAbilities(FuncEnum):
    # ꒰(˶• ᴗ •˶)꒱
    AXOLOTL_MORPH = "morph_into_salamander"


class EqualityTest(TestCase):
    def test_basic(self) -> None:
        self.assertEqual(  # Check underlying function correctness.
            AnimalAbilities.AXOLOTL_MORPH._get_function_for_value(),
            morph_into_salamander,
        )

    def test_call(self) -> None:
        # Should be too early to morph...
        self.assertFalse(AnimalAbilities.AXOLOTL_MORPH(how_soon=1))
        # Should've morphed yesterday!
        self.assertTrue(AnimalAbilities.AXOLOTL_MORPH(how_soon=-1))



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy

import numpy as np
from ax.adapter.base import DataLoaderConfig
from ax.adapter.data_utils import extract_experiment_data
from ax.adapter.transforms.search_space_to_choice import SearchSpaceToChoice
from ax.core.arm import Arm
from ax.core.experiment import Experiment
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)
from ax.core.search_space import SearchSpace
from ax.core.types import TParameterization
from ax.exceptions.core import DataRequiredError, UnsupportedError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_experiment_with_observations,
    get_robust_search_space,
)
from pandas.testing import assert_frame_equal, assert_series_equal


class SearchSpaceToChoiceTest(TestCase):
    def setUp(self) -> None:
        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "a", lower=1, upper=3, parameter_type=ParameterType.FLOAT
                ),
                ChoiceParameter(
                    "b", parameter_type=ParameterType.STRING, values=["a", "b", "c"]
                ),
            ]
        )
        parameterizations: list[TParameterization] = [
            {"a": 2, "b": "a"},
            {"a": 3, "b": "b"},
            {"a": 3, "b": "c"},
        ]
        experiment = get_experiment_with_observations(
            observations=[[1.0], [2.0], [3.0]],
            search_space=self.search_space,
            parameterizations=parameterizations,
        )
        self.experiment_data = extract_experiment_data(
            experiment=experiment, data_loader_config=DataLoaderConfig()
        )
        self.observation_features = [
            ObservationFeatures(parameters=p) for p in parameterizations
        ]
        self.signature_to_parameterization = {
            Arm(parameters=p).signature: p for p in parameterizations
        }
        self.transformed_features = [
            ObservationFeatures(
                parameters={"arms": Arm(parameters={"a": 2, "b": "a"}).signature}
            ),
            ObservationFeatures(
                parameters={"arms": Arm(parameters={"a": 3, "b": "b"}).signature}
            ),
            ObservationFeatures(
                parameters={"arms": Arm(parameters={"a": 3, "b": "c"}).signature}
            ),
        ]
        self.observations = [
            Observation(
                data=ObservationData([], np.array([]), np.empty((0, 0))), features=obsf
            )
            for obsf in self.observation_features
        ]
        self.t = SearchSpaceToChoice(
            search_space=self.search_space, experiment_data=self.experiment_data
        )
        self.t2 = SearchSpaceToChoice(
            search_space=self.search_space, observations=self.observations[:1]
        )
        self.t3 = SearchSpaceToChoice(
            search_space=self.search_space,
            observations=self.observations,
            config={"use_ordered": True},
        )

    def test_validation(self) -> None:
        # Test with no data.
        with self.assertRaisesRegex(DataRequiredError, "non-empty data"):
            SearchSpaceToChoice(
                search_space=self.search_space,
                observations=[],
            )
        # Test with empty experiment data.
        with self.assertRaisesRegex(DataRequiredError, "non-empty data"):
            SearchSpaceToChoice(
                search_space=self.search_space,
                experiment_data=extract_experiment_data(
                    experiment=Experiment(search_space=self.search_space),
                    data_loader_config=DataLoaderConfig(),
                ),
            )

        # Test error if there are fidelities.
        ss = SearchSpace(
            parameters=[
                RangeParameter(
                    "a",
                    lower=1,
                    upper=3,
                    parameter_type=ParameterType.FLOAT,
                    is_fidelity=True,
                    target_value=3,
                )
            ]
        )
        with self.assertRaisesRegex(ValueError, "fidelity"):
            SearchSpaceToChoice(
                search_space=ss,
                observations=self.observations,
            )

        # Test for error with robust search space.
        rss = get_robust_search_space()
        with self.assertRaisesRegex(
            UnsupportedError, "not supported for RobustSearchSpace"
        ):
            SearchSpaceToChoice(
                search_space=rss,
                observations=self.observations,
            )

    def test_TransformSearchSpace(self) -> None:
        ss2 = self.search_space.clone()
        ss2 = self.t.transform_search_space(ss2)
        self.assertEqual(len(ss2.parameters), 1)
        expected_parameter = ChoiceParameter(
            name="arms",
            parameter_type=ParameterType.STRING,
            values=list(self.t.signature_to_parameterization.keys()),
        )
        self.assertEqual(ss2.parameters.get("arms"), expected_parameter)

        # With use_ordered
        ss2 = self.search_space.clone()
        ss2 = self.t3.transform_search_space(ss2)
        self.assertEqual(len(ss2.parameters), 1)
        expected_parameter = ChoiceParameter(
            name="arms",
            parameter_type=ParameterType.STRING,
            values=list(self.t.signature_to_parameterization.keys()),
            is_ordered=True,
        )
        self.assertEqual(ss2.parameters.get("arms"), expected_parameter)

    def test_TransformSearchSpaceWithFixedParam(self) -> None:
        ss2 = self.search_space.clone()
        ss2 = self.t2.transform_search_space(ss2)
        self.assertEqual(len(ss2.parameters), 1)
        expected_parameter = FixedParameter(
            name="arms",
            parameter_type=ParameterType.STRING,
            value=list(self.t2.signature_to_parameterization.keys())[0],
        )
        self.assertEqual(ss2.parameters.get("arms"), expected_parameter)

    def test_TransformObservationFeatures(self) -> None:
        obs_ft2 = deepcopy(self.observation_features)
        obs_ft2 = self.t.transform_observation_features(obs_ft2)
        self.assertEqual(obs_ft2, self.transformed_features)
        obs_ft2 = self.t.untransform_observation_features(obs_ft2)
        self.assertEqual(obs_ft2, self.observation_features)

        # Testing transform empty parameters dict
        # Both transform and untransform should leave the param dict intact
        empty_obs_param = ObservationFeatures(parameters={}, trial_index=0)
        tsfm_empty_obs_param = self.t.transform_observation_features([empty_obs_param])[
            0
        ]
        self.assertEqual(tsfm_empty_obs_param, empty_obs_param)
        untsfm_empty_obs_param = self.t.untransform_observation_features(
            [tsfm_empty_obs_param]
        )[0]
        self.assertEqual(untsfm_empty_obs_param, empty_obs_param)

    def test_transform_experiment_data(self) -> None:
        # Empty data is returned unchanged.
        empty_data = extract_experiment_data(
            experiment=Experiment(search_space=self.search_space),
            data_loader_config=DataLoaderConfig(),
        )
        copy_empty_data = deepcopy(empty_data)
        transformed_data = self.t.transform_experiment_data(
            experiment_data=copy_empty_data
        )
        self.assertIs(copy_empty_data, transformed_data)
        self.assertEqual(transformed_data, empty_data)

        # Data is transformed to the signature.
        transformed_data = self.t.transform_experiment_data(
            experiment_data=deepcopy(self.experiment_data)
        )
        # Columns only include arms and metadata.
        self.assertEqual(set(transformed_data.arm_data), {"arms", "metadata"})
        # Metadata is unchanged.
        assert_series_equal(
            transformed_data.arm_data["metadata"],
            self.experiment_data.arm_data["metadata"],
        )
        # Arms are replaced by signatures.
        expected_arms = list(self.signature_to_parameterization)
        self.assertEqual(transformed_data.arm_data["arms"].tolist(), expected_arms)
        # Observation data is unchanged.
        assert_frame_equal(
            transformed_data.observation_data, self.experiment_data.observation_data
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.analysis.plotly.surface.contour import ContourPlot
from ax.analysis.plotly.surface.slice import SlicePlot

__all__ = ["ContourPlot", "SlicePlot"]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.core.experiment import Experiment
from pyre_extensions import none_throws

BASELINE_ARM_NAME = "baseline_arm"


def select_baseline_name_default_first_trial(
    experiment: Experiment, baseline_arm_name: str | None
) -> tuple[str, bool]:
    """
    Choose a baseline arm from arms on the experiment. Logic:
    1. If ``baseline_arm_name`` provided, validate that arm exists
       and return that arm name.
    2. If ``experiment.status_quo`` is set, return its arm name.
    3. If there is at least one trial on the experiment, use the
       first trial's first arm as the baseline.
    4. Error if 1-3 all don't apply.

    Returns:
        Tuple:
            baseline arm name (str)
            true when baseline selected from first arm of experiment (bool)
        raise ValueError if no valid baseline found
    """

    arms_dict = experiment.arms_by_name

    if baseline_arm_name:
        if baseline_arm_name not in arms_dict:
            raise ValueError(f"Arm by name {baseline_arm_name=} not found.")
        return baseline_arm_name, False

    if experiment.status_quo and none_throws(experiment.status_quo).name in arms_dict:
        baseline_arm_name = none_throws(experiment.status_quo).name
        return baseline_arm_name, False

    if (
        experiment.trials
        and experiment.trials[0].arms
        and experiment.trials[0].arms[0].name in arms_dict
    ):
        baseline_arm_name = experiment.trials[0].arms[0].name
        return baseline_arm_name, True

    else:
        raise ValueError("Could not find valid baseline arm.")



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from typing import Mapping, Sequence

import numpy as np
from ax.adapter.adapter_utils import array_to_observation_data, get_fixed_features
from ax.adapter.base import Adapter, DataLoaderConfig, GenResults
from ax.adapter.data_utils import ExperimentData
from ax.adapter.torch import (
    extract_objective_weights,
    extract_outcome_constraints,
    validate_transformed_optimization_config,
)
from ax.adapter.transforms.base import Transform
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.observation import ObservationData, ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.parameter import ChoiceParameter, FixedParameter
from ax.core.search_space import SearchSpace
from ax.core.types import TParamValueList
from ax.exceptions.core import DataRequiredError, UserInputError
from ax.generators.discrete_base import DiscreteGenerator
from ax.generators.types import TConfig


FIT_MODEL_ERROR = "Generator must be fit before {action}."


class DiscreteAdapter(Adapter):
    """An adapter for using generators based on discrete parameters.

    Requires that all parameters to have been transformed to ChoiceParameters.
    """

    def __init__(
        self,
        *,
        experiment: Experiment,
        generator: DiscreteGenerator,
        search_space: SearchSpace | None = None,
        data: Data | None = None,
        transforms: Sequence[type[Transform]] | None = None,
        transform_configs: Mapping[str, TConfig] | None = None,
        optimization_config: OptimizationConfig | None = None,
        expand_model_space: bool = True,
        fit_tracking_metrics: bool = True,
        fit_on_init: bool = True,
        data_loader_config: DataLoaderConfig | None = None,
        fit_out_of_design: bool | None = None,
        fit_abandoned: bool | None = None,
        fit_only_completed_map_metrics: bool | None = None,
    ) -> None:
        # These are set in _fit.
        self.parameters: list[str] = []
        self.outcomes: list[str] = []
        super().__init__(
            experiment=experiment,
            generator=generator,
            search_space=search_space,
            data=data,
            transforms=transforms,
            transform_configs=transform_configs,
            optimization_config=optimization_config,
            expand_model_space=expand_model_space,
            data_loader_config=data_loader_config,
            fit_out_of_design=fit_out_of_design,
            fit_abandoned=fit_abandoned,
            fit_tracking_metrics=fit_tracking_metrics,
            fit_on_init=fit_on_init,
            fit_only_completed_map_metrics=fit_only_completed_map_metrics,
        )
        # Re-assign for more precise typing.
        self.generator: DiscreteGenerator = generator

    def _fit(
        self,
        search_space: SearchSpace,
        experiment_data: ExperimentData,
    ) -> None:
        # Convert observations to arrays
        self.parameters = list(search_space.parameters.keys())
        if experiment_data.observation_data.empty:
            self.outcomes = []
        else:
            self.outcomes = sorted(experiment_data.metric_names)
        # Convert observations to arrays
        Xs_array, Ys_array, Yvars_array = self._convert_experiment_data(
            experiment_data=experiment_data,
            outcomes=self.outcomes,
            parameters=self.parameters,
        )
        # Extract parameter values
        parameter_values = _get_parameter_values(search_space, self.parameters)
        self.generator.fit(
            Xs=Xs_array,
            Ys=Ys_array,
            Yvars=Yvars_array,
            parameter_values=parameter_values,
            outcome_names=self.outcomes,
        )

    def _predict(
        self,
        observation_features: list[ObservationFeatures],
        use_posterior_predictive: bool = False,
    ) -> list[ObservationData]:
        # Convert observations to array
        X = [
            [of.parameters[param] for param in self.parameters]
            for of in observation_features
        ]
        f, cov = self.generator.predict(
            X=X, use_posterior_predictive=use_posterior_predictive
        )
        # Convert arrays to observations
        return array_to_observation_data(f=f, cov=cov, outcomes=self.outcomes)

    def _validate_gen_inputs(
        self,
        n: int,
        search_space: SearchSpace | None = None,
        optimization_config: OptimizationConfig | None = None,
        pending_observations: dict[str, list[ObservationFeatures]] | None = None,
        fixed_features: ObservationFeatures | None = None,
        model_gen_options: TConfig | None = None,
    ) -> None:
        """Validate inputs to `Adapter.gen`.

        Currently, this is only used to ensure that `n` is a positive integer or -1.
        """
        if n < 1 and n != -1:
            raise UserInputError(
                f"Attempted to generate n={n} points. Number of points to generate "
                "must be either a positive integer or -1."
            )

    def _gen(
        self,
        n: int,
        search_space: SearchSpace,
        pending_observations: dict[str, list[ObservationFeatures]],
        fixed_features: ObservationFeatures | None,
        model_gen_options: TConfig | None = None,
        optimization_config: OptimizationConfig | None = None,
    ) -> GenResults:
        """Generate new candidates according to search_space and
        optimization_config.

        Any outcome constraints should be transformed to no longer be relative, this
        can be achieved using either ``Relativize`` or ``Derelativize`` transforms.
        An error will be raised during validation if transform has not been applied.
        """
        # Validation
        if not self.parameters:
            raise ValueError(FIT_MODEL_ERROR.format(action="_gen"))
        # Extract parameter values
        parameter_values = _get_parameter_values(search_space, self.parameters)
        # Extract objective and outcome constraints
        if len(self.outcomes) == 0 or optimization_config is None:
            objective_weights = None
            outcome_constraints = None
        else:
            validate_transformed_optimization_config(optimization_config, self.outcomes)
            objective_weights = extract_objective_weights(
                objective=optimization_config.objective, outcomes=self.outcomes
            )
            outcome_constraints = extract_outcome_constraints(
                outcome_constraints=optimization_config.outcome_constraints,
                outcomes=self.outcomes,
            )

        # Get fixed features
        fixed_features_dict = get_fixed_features(
            fixed_features=fixed_features, param_names=self.parameters
        )

        # Pending observations
        if len(pending_observations) == 0:
            pending_array: list[list[TParamValueList]] | None = None
        else:
            pending_array = [[] for _ in self.outcomes]
            for metric_name, po_list in pending_observations.items():
                pending_array[self.outcomes.index(metric_name)] = [
                    [po.parameters[p] for p in self.parameters] for po in po_list
                ]

        # Generate the candidates
        X, w, gen_metadata = self.generator.gen(
            n=n,
            parameter_values=parameter_values,
            objective_weights=objective_weights,
            outcome_constraints=outcome_constraints,
            fixed_features=fixed_features_dict,  # pyre-ignore
            pending_observations=pending_array,
            model_gen_options=model_gen_options,
        )
        observation_features = [
            ObservationFeatures(parameters=dict(zip(self.parameters, x))) for x in X
        ]

        if "best_x" in gen_metadata:
            best_observation_features = ObservationFeatures(
                parameters=dict(zip(self.parameters, gen_metadata["best_x"]))
            )
        else:
            best_observation_features = None

        return GenResults(
            observation_features=observation_features,
            weights=w,
            gen_metadata=gen_metadata,
            best_observation_features=best_observation_features,
        )

    def _cross_validate(
        self,
        search_space: SearchSpace,
        cv_training_data: ExperimentData,
        cv_test_points: list[ObservationFeatures],
        use_posterior_predictive: bool = False,
    ) -> list[ObservationData]:
        """Make predictions at ``cv_test_points`` using only the data
        in ``cv_training_data``.
        """
        Xs_train, Ys_train, Yvars_train = self._convert_experiment_data(
            experiment_data=cv_training_data,
            outcomes=self.outcomes,
            parameters=self.parameters,
        )
        X_test = [
            [obsf.parameters[param] for param in self.parameters]
            for obsf in cv_test_points
        ]
        # Use the generator to do the cross validation
        f_test, cov_test = self.generator.cross_validate(
            Xs_train=Xs_train,
            Ys_train=Ys_train,
            Yvars_train=Yvars_train,
            X_test=X_test,
            use_posterior_predictive=use_posterior_predictive,
        )
        # Convert array back to ObservationData
        return array_to_observation_data(f=f_test, cov=cov_test, outcomes=self.outcomes)

    @classmethod
    def _convert_experiment_data(
        cls,
        experiment_data: ExperimentData,
        outcomes: list[str],
        parameters: list[str],
    ) -> tuple[list[list[TParamValueList]], list[list[float]], list[list[float]]]:
        """Convert experiment data to lists of Xs, Ys, and Yvars.
        Each element of the outer list contains the data for the corresponding outcome.
        """
        if len(outcomes) == 0:
            return [], [], []
        arm_data = experiment_data.arm_data
        obs_data = experiment_data.observation_data
        sems_df = obs_data["sem"]
        # Join mean & arms to align & repeat the arm rows if necessary.
        mean_and_params = obs_data["mean"].join(arm_data, how="left")
        params_np = mean_and_params[parameters].to_numpy()
        Xs: list[list[TParamValueList]] = []
        Ys: list[list[float]] = []
        Yvars: list[list[float]] = []
        for outcome in outcomes:
            if outcome not in mean_and_params:
                raise DataRequiredError(
                    f"Attempting to extract a data for {outcome=} but no data "
                    f"for {outcome} was found in the experiment data."
                )
            outcome_means = mean_and_params[outcome].to_numpy()
            to_keep = ~np.isnan(outcome_means)
            Xs.append(params_np[to_keep].tolist())
            Ys.append(outcome_means[to_keep].tolist())
            Yvars.append((sems_df[outcome][to_keep] ** 2).tolist())
        return Xs, Ys, Yvars


def _get_parameter_values(
    search_space: SearchSpace, param_names: list[str]
) -> list[TParamValueList]:
    """Extract parameter values from a search space of discrete parameters."""
    parameter_values: list[TParamValueList] = []
    for p_name in param_names:
        p = search_space.parameters[p_name]
        # Validation
        if isinstance(p, ChoiceParameter):
            # Set values
            parameter_values.append(p.values)
        elif isinstance(p, FixedParameter):
            parameter_values.append([p.value])
        else:
            raise ValueError(f"{p} not ChoiceParameter or FixedParameter")
    return parameter_values



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.benchmark.benchmark_metric import BenchmarkMetric
from ax.benchmark.benchmark_problem import (
    BenchmarkProblem,
    get_continuous_search_space,
    get_moo_opt_config,
    get_soo_opt_config,
)
from ax.benchmark.benchmark_test_functions.botorch_test import BoTorchTestFunction
from ax.core.objective import MultiObjective, Objective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.outcome_constraint import OutcomeConstraint
from ax.core.search_space import SearchSpace
from ax.core.types import ComparisonOp
from ax.utils.common.testutils import TestCase
from botorch.test_functions.synthetic import Branin
from pyre_extensions import assert_is_instance


class TestBenchmarkProblem(TestCase):
    def setUp(self) -> None:
        # Print full output, so that any differences in 'repr' output are shown
        self.maxDiff = None
        super().setUp()

    def test_mismatch_of_names_on_test_function_and_opt_config_raises(self) -> None:
        objectives = [
            Objective(metric=BenchmarkMetric(name, lower_is_better=True))
            for name in ["Branin", "Currin"]
        ]
        test_function = BoTorchTestFunction(
            botorch_problem=Branin(), outcome_names=["Branin"]
        )
        opt_config = MultiObjectiveOptimizationConfig(
            objective=MultiObjective(objectives)
        )
        with self.assertRaisesRegex(
            ValueError,
            "The following objectives are defined on "
            "`optimization_config` but not included in "
            "`runner.test_function.outcome_names`: {'Currin'}.",
        ):
            BenchmarkProblem(
                name="foo",
                optimization_config=opt_config,
                num_trials=1,
                optimal_value=1.0,
                search_space=SearchSpace(parameters=[]),
                test_function=test_function,
                baseline_value=0.0,
            )

        opt_config = OptimizationConfig(
            objective=objectives[0],
            outcome_constraints=[
                OutcomeConstraint(
                    BenchmarkMetric("c", lower_is_better=False),
                    ComparisonOp.LEQ,
                    0.0,
                )
            ],
        )
        with self.assertRaisesRegex(
            ValueError,
            "The following constraints are defined on "
            "`optimization_config` but not included in "
            "`runner.test_function.outcome_names`: {'c'}.",
        ):
            BenchmarkProblem(
                name="foo",
                optimization_config=opt_config,
                num_trials=1,
                optimal_value=0.0,
                search_space=SearchSpace(parameters=[]),
                test_function=test_function,
                baseline_value=1.0,
            )

    def test_get_soo_opt_config(self) -> None:
        opt_config = get_soo_opt_config(
            outcome_names=["foo", "bar"],
            lower_is_better=False,
            observe_noise_sd=True,
        )
        self.assertIsInstance(opt_config.objective, Objective)
        self.assertEqual(len(opt_config.outcome_constraints), 1)
        objective_metric = assert_is_instance(
            opt_config.objective.metric, BenchmarkMetric
        )
        self.assertEqual(objective_metric.name, "foo")
        self.assertEqual(objective_metric.observe_noise_sd, True)
        self.assertEqual(objective_metric.lower_is_better, False)
        constraint_metric = assert_is_instance(
            opt_config.outcome_constraints[0].metric, BenchmarkMetric
        )
        self.assertEqual(constraint_metric.name, "bar")
        self.assertEqual(constraint_metric.observe_noise_sd, True)
        self.assertEqual(constraint_metric.lower_is_better, False)

    def test_get_moo_opt_config(self) -> None:
        opt_config = get_moo_opt_config(
            outcome_names=["foo", "bar", "baz", "pony"],
            ref_point=[3.0, 4.0],
            lower_is_better=False,
            observe_noise_sd=True,
            num_constraints=2,
        )
        self.assertEqual(len(opt_config.objective.metrics), 2)
        self.assertEqual(len(opt_config.outcome_constraints), 2)
        self.assertEqual(opt_config.objective_thresholds[0].bound, 3.0)
        objective_metrics = [
            assert_is_instance(metric, BenchmarkMetric)
            for metric in opt_config.objective.metrics
        ]
        self.assertEqual(len(objective_metrics), 2)
        self.assertEqual(objective_metrics[0].name, "foo")
        self.assertEqual(objective_metrics[0].observe_noise_sd, True)
        self.assertEqual(objective_metrics[0].lower_is_better, False)
        self.assertEqual(objective_metrics[1].name, "bar")
        # Check constraints
        self.assertEqual(len(opt_config.outcome_constraints), 2)
        constraint_metrics = [
            assert_is_instance(constraint.metric, BenchmarkMetric)
            for constraint in opt_config.outcome_constraints
        ]
        self.assertEqual(constraint_metrics[0].name, "baz")
        self.assertEqual(constraint_metrics[0].observe_noise_sd, True)
        self.assertEqual(constraint_metrics[0].lower_is_better, False)

    def test_get_continuous_search_space(self) -> None:
        bounds = [(0.0, 1.0), (2.0, 3.0)]
        with self.subTest("Dummy parameters not specified"):
            search_space = get_continuous_search_space(bounds=bounds)
            self.assertEqual(len(search_space.parameters), 2)
            self.assertEqual(
                len(search_space.parameters), len(search_space.range_parameters)
            )
            self.assertEqual({"x0", "x1"}, search_space.parameters.keys())

        with self.subTest("Dummy parameters specified"):
            search_space = get_continuous_search_space(
                bounds=bounds, n_dummy_dimensions=2
            )
            self.assertEqual(len(search_space.parameters), 4)
            self.assertEqual(
                len(search_space.parameters), len(search_space.range_parameters)
            )
            self.assertEqual(
                {"x0", "x1", "embedding_dummy_0", "embedding_dummy_1"},
                set(search_space.parameters.keys()),
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from typing import Iterable, Sequence

from ax.adapter.base import Adapter

from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCard
from ax.core.experiment import Experiment
from ax.core.trial_status import TrialStatus
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from pyre_extensions import override


class Summary(Analysis):
    """
    High-level summary of the Experiment with one row per arm. Any values missing at
    compute time will be represented as None. Columns where every value is None will
    be omitted by default.

    The DataFrame computed will contain one row per arm and the following columns:
        - trial_index: The trial index of the arm
        - arm_name: The name of the arm
        - trial_status: The status of the trial (e.g. RUNNING, SUCCEDED, FAILED)
        - failure_reason: The reason for the failure, if applicable
        - generation_node: The name of the ``GenerationNode`` that generated the arm
        - **METADATA: Any metadata associated with the trial, as specified by the
            Experiment's runner.run_metadata_report_keys field
        - **METRIC_NAME: The observed mean of the metric specified, for each metric
        - **PARAMETER_NAME: The parameter value for the arm, for each parameter
     Args:
        trial_indices: If specified, only include these trial indices.
        trial_status: If specified, only include trials with this status.
        omit_empty_columns: If True, omit columns where every value is None.
    """

    def __init__(
        self,
        trial_indices: Iterable[int] | None = None,
        trial_statuses: Sequence[TrialStatus] | None = None,
        omit_empty_columns: bool = True,
    ) -> None:
        self.trial_indices = trial_indices
        self.trial_statuses = trial_statuses
        self.omit_empty_columns = omit_empty_columns

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCard:
        if experiment is None:
            raise UserInputError("`Summary` analysis requires an `Experiment` input")

        return self._create_analysis_card(
            title=(
                "Summary for "
                f"{experiment.name if experiment.has_name else 'Experiment'}"
            ),
            subtitle="High-level summary of the `Trial`-s in this `Experiment`",
            df=experiment.to_df(
                trial_indices=self.trial_indices,
                omit_empty_columns=self.omit_empty_columns,
                trial_statuses=self.trial_statuses,
            ),
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest.mock import patch

import torch
from ax.adapter.registry import Generators
from ax.adapter.transforms.choice_encode import OrderedChoiceToIntegerRange
from ax.generators.torch.botorch_modular.optimizer_defaults import (
    BATCH_LIMIT,
    INIT_BATCH_LIMIT,
)
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment
from ax.utils.testing.mock import mock_botorch_optimize_context_manager
from botorch.acquisition.analytic import ExpectedImprovement, PosteriorMean
from botorch.generation.gen import gen_candidates_scipy
from botorch.optim.optimize_mixed import generate_starting_points
from botorch.utils.testing import MockAcquisitionFunction, skip_if_import_error
from pyro.infer import MCMC


class TestMock(TestCase):
    def test_no_mocks_called(self) -> None:
        # Should raise by default if no mocks are called.
        with self.assertRaisesRegex(AssertionError, "No mocks were called"):
            with mock_botorch_optimize_context_manager():
                pass
        # Doesn't raise when force=True.
        with mock_botorch_optimize_context_manager(force=True):
            pass

    def test_botorch_mocks(self) -> None:
        # Should not raise when BoTorch mocks are called.
        with mock_botorch_optimize_context_manager():
            gen_candidates_scipy(
                initial_conditions=torch.tensor([[0.0]]),
                acquisition_function=MockAcquisitionFunction(),  # pyre-ignore [6]
            )

    def test_fully_bayesian_mocks(self) -> None:
        experiment = get_branin_experiment(with_completed_batch=True)
        with patch("botorch.fit.MCMC", wraps=MCMC) as mock_mcmc:
            with mock_botorch_optimize_context_manager():
                Generators.SAASBO(experiment=experiment, data=experiment.lookup_data())
        mock_mcmc.assert_called_once()
        kwargs = mock_mcmc.call_args.kwargs
        self.assertEqual(kwargs["num_samples"], 16)
        self.assertEqual(kwargs["warmup_steps"], 0)

    def test_mixed_optimizer_mocks(self) -> None:
        experiment = get_branin_experiment(
            with_completed_batch=True, with_choice_parameter=True
        )
        with patch(
            "botorch.optim.optimize_mixed.generate_starting_points",
            wraps=generate_starting_points,
        ) as mock_gen:
            with mock_botorch_optimize_context_manager():
                Generators.BOTORCH_MODULAR(
                    experiment=experiment,
                    data=experiment.lookup_data(),
                    transforms=[OrderedChoiceToIntegerRange],
                ).gen(n=1)
        mock_gen.assert_called_once()
        opt_inputs = mock_gen.call_args.kwargs["opt_inputs"]
        self.assertEqual(opt_inputs.raw_samples, 2)
        self.assertEqual(opt_inputs.num_restarts, 1)
        self.assertEqual(
            opt_inputs.options,
            {
                "init_batch_limit": INIT_BATCH_LIMIT,
                "batch_limit": BATCH_LIMIT,
                "maxiter_alternating": 1,
                "maxiter_continuous": 1,
                "maxiter_init": 1,
                "maxiter_discrete": 1,
            },
        )

    @skip_if_import_error
    def test_optimize_with_nsgaii_mocks(self) -> None:
        from pymoo.algorithms.moo.nsga2 import NSGA2
        from pymoo.optimize import minimize

        experiment = get_branin_experiment(with_completed_batch=True)
        with patch(
            "botorch.utils.multi_objective.optimize.minimize",
            wraps=minimize,
        ) as mock_minimize, patch(
            "botorch.utils.multi_objective.optimize.NSGA2",
            wraps=NSGA2,
        ) as mock_nsgaii:
            with mock_botorch_optimize_context_manager():
                Generators.BOTORCH_MODULAR(
                    experiment=experiment,
                    data=experiment.lookup_data(),
                    botorch_acqf_classes_with_options=[
                        (PosteriorMean, {}),
                        (ExpectedImprovement, {}),
                    ],
                ).gen(n=1)
        mock_minimize.assert_called_once()
        mock_nsgaii.assert_called_once()
        self.assertEqual(mock_nsgaii.call_args.kwargs["pop_size"], 10)
        self.assertEqual(mock_minimize.call_args.kwargs["termination"].n_max_gen, 1)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Optional, TYPE_CHECKING

from ax.adapter.data_utils import ExperimentData
from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.utils import construct_new_search_space
from ax.core.observation import Observation
from ax.core.parameter import FixedParameter, Parameter, RangeParameter
from ax.core.search_space import SearchSpace
from ax.generators.types import TConfig

if TYPE_CHECKING:
    # import as module to make sphinx-autodoc-typehints happy
    from ax import adapter as adapter_module  # noqa F401


class FixedToTunable(Transform):
    """Convert any FixedParameter to a RangeParameter if the SearchSpace that
    is passed on initialization has a RangeParameter of the same name.

    TODO: Add support for ChoiceParameters.
    """

    no_op_for_experiment_data = True

    def __init__(
        self,
        search_space: SearchSpace | None = None,
        observations: list[Observation] | None = None,
        experiment_data: ExperimentData | None = None,
        adapter: Optional["adapter_module.base.Adapter"] = None,
        config: TConfig | None = None,
    ) -> None:
        assert search_space is not None, "FixedToTunable requires search space"
        super().__init__(
            search_space=search_space,
            observations=observations,
            experiment_data=experiment_data,
            adapter=adapter,
            config=config,
        )
        # Store search space. In transfer learning, this will be the joint search
        # space that contains range parameters when one experiment has a fixed
        # parameter and another experiment has a range parameter of the same name.
        self.search_space: SearchSpace = search_space

    def _transform_search_space(self, search_space: SearchSpace) -> SearchSpace:
        transformed_parameters: dict[str, Parameter] = {}
        for p_name, p in search_space.parameters.items():
            if isinstance(p, FixedParameter) and isinstance(
                target_p := self.search_space.parameters[p_name], RangeParameter
            ):
                transformed_parameters[p_name] = target_p.clone()
            else:
                transformed_parameters[p.name] = p

        return construct_new_search_space(
            search_space=search_space,
            parameters=list(transformed_parameters.values()),
            # the target search space cannot have constraints on a fixed
            # so the parameter constraints will not have changed here.
            parameter_constraints=search_space.parameter_constraints,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Collection
from dataclasses import dataclass, field, InitVar
from pathlib import Path

import pandas as pd

import torch
from ax.benchmark.problems.data import AbstractParquetDataLoader
from ax.benchmark.problems.surrogate.lcbench.utils import (
    DEFAULT_METRIC_NAME,
    get_lcbench_log_scale_parameter_names,
    get_lcbench_parameter_names,
)

DATASET_NAMES = [
    "APSFailure",
    "Amazon_employee_access",
    "Australian",
    "Fashion-MNIST",
    "KDDCup09_appetency",
    "MiniBooNE",
    "adult",
    "airlines",
    "albert",
    "bank-marketing",
    "blood-transfusion-service-center",
    "car",
    "christine",
    "cnae-9",
    "connect-4",
    "covertype",
    "credit-g",
    "dionis",
    "fabert",
    "helena",
    "higgs",
    "jannis",
    "jasmine",
    "jungle_chess_2pcs_raw_endgame_complete",
    "kc1",
    "kr-vs-kp",
    "mfeat-factors",
    "nomao",
    "numerai28.6",
    "phoneme",
    "segment",
    "shuttle",
    "sylvine",
    "vehicle",
    "volkert",
]


class LCBenchDataLoader(AbstractParquetDataLoader):
    def __init__(
        self,
        dataset_name: str,
        stem: str,
        cache_dir: Path | None = None,
    ) -> None:
        super().__init__(
            benchmark_name="LCBenchLite",
            dataset_name=dataset_name,
            stem=stem,
            cache_dir=cache_dir,
        )

    @property
    def url(self) -> str:
        """
        URL to the GZIP compressed parquet files for the 35 datasets from LCBench.
        These files were created by splitting the massive JSON dump of LCBench into
        datasets, then further into config info, learning curve metrics, and final
        results, and subsequently saving them to an efficient Parquet format,
        compressed with GZIP, and finally uploading them to address.
        """

        return (
            "https://raw.githubusercontent.com/ltiao/"
            f"{self.benchmark_name}/main/{self.dataset_name}/{self.filename}"
        )


@dataclass(kw_only=True)
class LCBenchData:
    """
    Args:
        parameter_df: DataFrame with columns corresponding to the names of the
            parameters in get_lcbench_parameter_names().
        metric_series: Series of metric values with index names "trial" and "epoch".
        timestamp_series: Series of timestamps with index name "trial".
    """

    parameter_df: pd.DataFrame
    metric_series: pd.Series
    timestamp_series: pd.Series

    runtime_series: pd.Series = field(init=False)
    # pyre-ignore [16]: Pyre doesn't understand InitVars.
    runtime_fillna: InitVar[bool] = False
    # pyre-ignore [16]: Pyre doesn't understand InitVars.
    log_scale_parameter_names: InitVar[Collection[str] | None] = None
    dtype: torch.dtype = torch.double
    device: torch.device | None = None

    def __post_init__(
        self,
        runtime_fillna: bool,
        log_scale_parameter_names: Collection[str] | None,
    ) -> None:
        self.timestamp_series.name = "timestamp"

        self.runtime_series = self._get_runtime_series(fillna=runtime_fillna)
        self.runtime_series.name = "runtimes"

        parameter_names = get_lcbench_parameter_names()
        if log_scale_parameter_names is None:
            log_scale_parameter_names = get_lcbench_log_scale_parameter_names()

        if len(log_scale_parameter_names) > 0:
            if unrecognized_param_set := (
                set(log_scale_parameter_names) - set(parameter_names)
            ):
                raise ValueError(f"Unrecognized columns: {unrecognized_param_set}")
            self.parameter_df[log_scale_parameter_names] = self.parameter_df[
                log_scale_parameter_names
            ].transform("log")

        self.parameter_df = self.parameter_df[parameter_names]

    @staticmethod
    def _unstack_by_epoch(series: pd.Series) -> pd.DataFrame:
        # unstack by epoch and truncate 52 epochs [0, ..., 51]
        # to 50 epochs [1, ..., 50]
        return series.unstack(level="epoch").iloc[:, 1:-1]

    def _get_runtime_series(self, fillna: bool) -> pd.Series:
        # timestamp (in secs) at every epoch, grouped by trial
        timestamps_grouped = self.timestamp_series.groupby(level="trial")

        # runtime (in secs) of training each incremental epoch
        runtime_series = timestamps_grouped.diff(periods=1)  # first element is NaN
        if fillna:
            runtime_series.fillna(timestamps_grouped.head(n=1), inplace=True)

        return runtime_series

    def _to_tensor(
        self,
        x: pd.DataFrame | pd.Series,
    ) -> torch.Tensor:
        return torch.from_numpy(x.values).to(dtype=self.dtype, device=self.device)

    @property
    def metric_df(self) -> pd.DataFrame:
        return self._unstack_by_epoch(self.metric_series)

    @property
    def runtime_df(self) -> pd.DataFrame:
        return self._unstack_by_epoch(self.runtime_series)

    @property
    def average_runtime_series(self) -> pd.Series:
        # take average runtime over epochs (N6231489 shows runtime is
        # mostly constant across epochs, as one'd expect)
        return self.runtime_series.groupby(level="trial").mean()

    @property
    def parameters(self) -> torch.Tensor:
        return self._to_tensor(self.parameter_df)

    @property
    def metrics(self) -> torch.Tensor:
        return self._to_tensor(self.metric_df)

    @property
    def runtimes(self) -> torch.Tensor:
        return self._to_tensor(self.runtime_df)

    @property
    def average_runtimes(self) -> torch.Tensor:
        return self._to_tensor(self.average_runtime_series)


def load_lcbench_data(
    dataset_name: str,
    metric_name: str = DEFAULT_METRIC_NAME,
    log_scale_parameter_names: Collection[str] | None = None,
    dtype: torch.dtype = torch.double,
    device: torch.device | None = None,
) -> LCBenchData:
    if dataset_name not in DATASET_NAMES:
        raise ValueError(
            f"Invalid dataset {dataset_name}. Valid datasets: {DATASET_NAMES}"
        )

    parameter_df = LCBenchDataLoader(dataset_name, stem="config").load()
    metrics_df = LCBenchDataLoader(dataset_name, stem="metrics").load()

    return LCBenchData(
        parameter_df=parameter_df,
        metric_series=metrics_df[metric_name],
        timestamp_series=metrics_df["time"],
        log_scale_parameter_names=log_scale_parameter_names,
        dtype=dtype,
        device=device,
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest.mock import patch

import numpy as np
from ax.benchmark.problems.surrogate.lcbench.early_stopping import (
    _get_lcbench_early_stopping_benchmark_problem,
    get_lcbench_early_stopping_benchmark_problem,
    LearningCurveBenchmarkTestFunction,
    OPTIMAL_VALUES,
    RUNTIME_MULTIPLIERS,
)
from ax.benchmark.problems.surrogate.lcbench.utils import (
    BASELINE_VALUES,
    DEFAULT_METRIC_NAME,
)
from ax.benchmark.testing.benchmark_stubs import get_mock_lcbench_data
from ax.utils.common.testutils import TestCase
from pyre_extensions import assert_is_instance, none_throws


class TestEarlyStoppingProblem(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.early_stopping_path = (
            _get_lcbench_early_stopping_benchmark_problem.__module__
        )

    def test_get_lcbench_early_stopping_problem(self) -> None:
        # Just test one problem for speed. We are mocking out the data load
        # anyway, so there is nothing to distinguish these problems from each
        # other

        num_trials = 4
        dataset_name = "credit-g"

        with (
            patch(
                f"{self.early_stopping_path}.load_lcbench_data",
                return_value=get_mock_lcbench_data(),
            ) as mock_load_lcbench_data,
            patch(
                # Fitting a surrogate won't work with this small synthetic data
                f"{self.early_stopping_path}._create_surrogate_regressor"
            ) as mock_create_surrogate_regressor,
        ):
            problem = get_lcbench_early_stopping_benchmark_problem(
                dataset_name=dataset_name,
                num_trials=num_trials,
                constant_step_runtime=True,
            )

        mock_load_lcbench_data.assert_called_once()
        mock_load_lcbench_data_kwargs = mock_load_lcbench_data.call_args.kwargs
        self.assertEqual(mock_load_lcbench_data_kwargs["dataset_name"], dataset_name)
        create_surrogate_regressor_call_args = (
            mock_create_surrogate_regressor.call_args_list
        )
        self.assertEqual(len(create_surrogate_regressor_call_args), 2)
        self.assertEqual(
            problem.optimization_config.objective.metric.name, DEFAULT_METRIC_NAME
        )
        self.assertIsNone(problem.step_runtime_function)
        self.assertEqual(problem.optimal_value, OPTIMAL_VALUES[dataset_name])
        self.assertEqual(problem.baseline_value, BASELINE_VALUES[dataset_name])

    def test_step_scaling(self) -> None:
        dataset_name = "car"
        with (
            patch(
                f"{self.early_stopping_path}.load_lcbench_data",
                return_value=get_mock_lcbench_data(),
            ),
            patch(
                # Fitting a surrogate won't work with this small synthetic data
                f"{self.early_stopping_path}._create_surrogate_regressor"
            ),
        ):
            problem = get_lcbench_early_stopping_benchmark_problem(
                dataset_name=dataset_name,
            )

        predicted_runtime = 1234.5
        test_function = assert_is_instance(
            problem.test_function, LearningCurveBenchmarkTestFunction
        )
        # pyre-fixme[8]: Incompatible attribute type -- not a bound method
        test_function.runtime_surrogate.predict = lambda X: np.array(
            [predicted_runtime]
        )
        self.assertEqual(
            none_throws(problem.step_runtime_function)(params={"param": 0}),
            predicted_runtime * RUNTIME_MULTIPLIERS[dataset_name],
        )
